/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.core;

import junit.framework.TestCase;
import pl.chyla.wyjec.testutils.ArrayTestingUtils;


/**
 * Tests for {@link Type} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestType extends TestCase {

    /**
     * Empty default constructor
     */
    public TestType() { }

    /**
     * Returns all publicly available {@link Type} instances.
     *
     * @return new array with all {@link Type} instances.
     */
    public static Type[] getAllTypes() {
        return new Type[] {Type.NONE, Type.BOOLEAN, Type.INTEGER, Type.STRING, Type.FLOAT};
    }

    /**
     * Returns all publicly available {@link Type} instances except {@link Type#NONE}.
     *
     * @return new array with all {@link Type} instances representing real types.
     */
    public static Type[] getAllRealTypes() {
        return new Type[] {Type.BOOLEAN, Type.INTEGER, Type.STRING, Type.FLOAT};
    }

    /**
     * Tests whether after encoding a type with {@link Type#encodedType} and decoding it later
     * with {@link Type#decodeType(int)} results with the original type. 
     */
    public void testDecodesType() {
        Type[] allTypes = getAllTypes();
        for (int i = 0; i < allTypes.length; i++) {
            assertSame(allTypes[i], Type.decodeType(allTypes[i].getEncodedType()));
        }
    }

    /**
     * Tests whether none of publicly available {@link Type} instances is equal to another one.
     */
    public void testTypesUniqueness() {
        assertTrue(ArrayTestingUtils.checkUniqueElements(getAllTypes()));
    }

    /**
     * Tests whether none of encoded forms of types is equal to another one.
     */
    public void testEncodedTypesUniqueness() {
        Type[] allTypes = getAllTypes();
        Integer[] encodedTypes = new Integer[allTypes.length];
        for (int i = 0; i < allTypes.length; i++) {
            encodedTypes[i] = new Integer(allTypes[i].getEncodedType());
        }
        assertTrue(ArrayTestingUtils.checkUniqueElements(encodedTypes));
    }

    /**
     * Tests {@link Type#decodeType(int)} with invalid <code>encodedType</code> argument.
     * TODO: don't use hard-coded numbers
     */
    public void testDecodeTypeInvalidArg() {
        try {
            Type.decodeType(-1);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
        try {
            Type.decodeType(Integer.MIN_VALUE);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
    }

}
