/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.core;

import junit.framework.TestCase;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;
import pl.chyla.wyjec.utils.SimpleCalendar;
import pl.chyla.wyjec.utils.TestRawString;


/**
 * Tests for {@link Value} class.
 * TODO: tests for equal() and hashCode() (see ReadableRawString)
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestValue extends TestCase {

    /**
     * Empty default constructor
     */
    public TestValue() { }

    /**
     * Creates new {@link Value} instance and sets it to boolean value based on the parameter.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    public static final Value makeBoolean(boolean value) {
        Value result = new Value(0);
        result.setBoolean(value);
        return result;
    }

    /**
     * Creates new {@link Value} instance and sets it to integer value based on the parameter.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    public static Value makeInteger(long value) {
        Value result = new Value(0);
        result.setInteger(value);
        return result;
    }

    /**
     * Creates new {@link Value} instance and sets it to string value based on the parameter.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    public static Value makeString(String value) {
        Value result = new Value(0);
        result.setString(new RawString(value));
        return result;
    }

    /**
     * Builds an array of {@link Value} instances bound to various {@link RawString} instances
     * with values equal to given string. {@link RawString} instances contained in these values
     * will differ with size of storage and its layout, but will have the same string value.
     *
     * @param value value to build {@link Value} instances for
     * @return a list of generated instances
     */
    public static Value[] makeStringVariants(String value) {
        RawString[] rawValues = TestRawString.buildRawStringVariants(value);
        Value[] values = new Value[rawValues.length];
        for (int i = 0; i < values.length; i++) {
            values[i] = new Value(0);
            values[i].setString(rawValues[i]);
        }
        return values;
    }

    /**
     * Creates new {@link Value} instance and sets it to float value based on the parameter.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    public static Value makeFloat(double value) {
        Value result = new Value(0);
        result.setFloat(value);
        return result;
    }

    /**
     * Creates new {@link Value} instance and sets it to date value based on the parameter.
     *
     * @param value value to embed in created instance
     * @return newly created {@link Value} instance
     */
    public static Value makeDate(long value) {
        Value result = new Value(0);
        result.setDate(value);
        return result;
    }

    /**
     * Creates new {@link Value} instance and sets it to date value specified by given
     * parameters.
     *
     * @param year year number
     * @param month 1-based month of the year
     * @param day 1-based day of the month
     * @param hour 0-based hour of the day
     * @param minute 0-based minute of the hour
     * @param second 0-based second of the minute
     * @return newly created {@link Value} instance
     */
    public static Value makeDate(
        int year, int month, int day, int hour, int minute, int second) {
        SimpleCalendar cal = new SimpleCalendar();
        cal.setDate(year, month - 1, day, hour, minute, second);
        Value result = new Value(0);
        result.setDate(cal.getTime().getTime());
        return result;
    }

    /**
     * Tests {@link Value#Value(int)} constructor.
     */
    public void testConstructor() {
        new Value(0);
        new Value(100);
    }

    /**
     * Tests {@link Value#Value(int)} constructor with invalid <code>maxStringLength</code>
     * argument.
     */
    public void testConstructorInvalidArgs() {
        try {
            new Value(-1);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
        try {
            new Value(Integer.MIN_VALUE);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
    }

    /**
     * Tests {@link Value#setBoolean(boolean)} and {@link Value#getBoolean()} methods. Also
     * checks if {@link Value#getType()} returns valid type for boolean values.
     */
    public void testSetGetBoolean() {
        Value v = new Value(0);
        v.setBoolean(false);
        assertSame(v.getType(), Type.BOOLEAN);
        assertEquals(v.getBoolean(), false);
        v.setBoolean(true);
        assertSame(v.getType(), Type.BOOLEAN);
        assertEquals(v.getBoolean(), true);
    }

    /**
     * Tests {@link Value#setInteger(long)} and {@link Value#getInteger()} methods. Also checks
     * if {@link Value#getType()} returns valid type for integer values.
     */
    public void testSetGetInteger() {
        Value v = new Value(0);
        v.setInteger(0L);
        assertSame(v.getType(), Type.INTEGER);
        assertEquals(v.getInteger(), 0L);
        v.setInteger(123456L);
        assertSame(v.getType(), Type.INTEGER);
        assertEquals(v.getInteger(), 123456L);
        v.setInteger(Long.MIN_VALUE);
        assertSame(v.getType(), Type.INTEGER);
        assertEquals(v.getInteger(), Long.MIN_VALUE);
    }

    /**
     * Tests {@link Value#setString(ReadableRawString)} and {@link Value#getString()} methods.
     * Also checks if {@link Value#getType()} returns valid type for string values.
     */
    public void testSetGetString() {
        Value v = new Value(0);
        RawString rs = new RawString("abc");
        v.setString(rs);
        assertSame(v.getType(), Type.STRING);
        assertEquals(v.getString(), rs);
        assertNotSame(v.getString(), rs);
    }

    /**
     * Tests {@link Value#setFloat(double)} and {@link Value#getFloat()} methods. Also checks if
     * {@link Value#getType()} returns valid type for float values.
     */
    public void testSetGetFloat() {
        Value v = new Value(0);
        v.setFloat(0.0);
        assertSame(v.getType(), Type.FLOAT);
        assertEquals(v.getFloat(), 0.0, 0.0);
        v.setFloat(123456.5);
        assertSame(v.getType(), Type.FLOAT);
        assertEquals(v.getFloat(), 123456.5, 0.0);
        v.setFloat(Double.MAX_VALUE);
        assertSame(v.getType(), Type.FLOAT);
        assertEquals(v.getFloat(), Double.MAX_VALUE, 0.0);
    }

    /**
     * Tests {@link Value#setDate(long)} and {@link Value#getDate()} methods. Also checks if
     * {@link Value#getType()} returns valid type for date values.
     */
    public void testSetGetDate() {
        Value v = new Value(0);
        v.setDate(0L);
        assertSame(v.getType(), Type.DATE);
        assertEquals(v.getDate(), 0L);
        SimpleCalendar cal = new SimpleCalendar();
        cal.setDate(1977, 2, 9);
        v.setDate(cal.getTimeInMillis());
        assertSame(v.getType(), Type.DATE);
        assertEquals(v.getDate(), cal.getTimeInMillis());
    }

    /**
     * Tests {@link Value#set(Value)} method.
     */
    public void testSet() {
        Value v1 = new Value(20);
        Value v2 = new Value(0);
        v2.set(v1);
        assertSame(v1.getType(), v2.getType());
        v1.setBoolean(true);
        v2.set(v1);
        assertSame(v2.getType(), Type.BOOLEAN);
        assertEquals(v1.getBoolean(), v2.getBoolean());
        v1.setInteger(12345L);
        v2.set(v1);
        assertSame(v2.getType(), Type.INTEGER);
        assertEquals(v1.getInteger(), v2.getInteger());
        v1.resetString();
        v1.getString().append("abc");
        v2.set(v1);
        assertSame(v2.getType(), Type.STRING);
        assertEquals(v1.getString(), v2.getString());
        assertSame(v1.getString().getStorage(), v2.getString().getStorage());
        v1.resetString();
        assertEquals(v2.getString().toString(), "abc");
        v1.setFloat(12345.5);
        v2.set(v1);
        assertSame(v2.getType(), Type.FLOAT);
        assertEquals(v1.getFloat(), v2.getFloat(), 0.0);
        SimpleCalendar cal = new SimpleCalendar();
        cal.setDate(1977, 2, 9);
        v1.setDate(cal.getTimeInMillis());
        v2.set(v1);
        assertSame(v2.getType(), Type.DATE);
        assertEquals(v1.getDate(), v2.getDate());
    }

    /**
     * Tests {@link Value#resetString()} method.
     */
    public void testResetString() {
        Value v = new Value(3);
        v.resetString();
        assertEquals(v.getString().getLength(), 0);
        v.getString().append("ab");
        v.resetString();
        assertEquals(v.getString().getLength(), 0);
        v.setString(new RawString("abc"));
        v.resetString();
        assertEquals(v.getString().getLength(), 0);
    }

    /**
     * Tests {@link Value#toString()} method. Verifies whether the method returns different
     * non-empty strings for a few different values and whether it's deterministic (returns the
     * same string for the same value). The actual strings are not verified as their format is
     * not documented.
     */
    public void testToString() {
        boolean booleanValue = false;
        long integerValue = 12345L;
        RawString stringValue = new RawString("test");
        double floatValue = 12345.0;
        SimpleCalendar cal = new SimpleCalendar();
        cal.setDate(1977, 2, 9);
        long dateValue = cal.getTimeInMillis();
        Value v = new Value(0);
        String s1 = v.toString();
        v.setBoolean(booleanValue);
        String s2 = v.toString();
        v.setBoolean(!booleanValue);
        String s3 = v.toString();
        v.setInteger(integerValue);
        String s4 = v.toString();
        v.setString(stringValue);
        String s5 = v.toString();
        v.setFloat(floatValue);
        String s6 = v.toString();
        v.setDate(dateValue);
        String s7 = v.toString();
        v = new Value(0);
        assertEquals(v.toString(), s1);
        v.setBoolean(booleanValue);
        assertEquals(v.toString(), s2);
        v.setBoolean(!booleanValue);
        assertEquals(v.toString(), s3);
        v.setInteger(integerValue);
        assertEquals(v.toString(), s4);
        v.setString(stringValue);
        assertEquals(v.toString(), s5);
        v.setFloat(floatValue);
        assertEquals(v.toString(), s6);
        v.setDate(dateValue);
        assertEquals(v.toString(), s7);
    }

    /**
     * Generates a sequence of {@link Value} instances such that for some instances there're
     * others in the sequence with equal values.
     *
     * @return array of sample {@link Value} instances
     */
    private Value[] generateSampleValues() {
        return new Value[] {
            makeBoolean(false), makeBoolean(false), makeBoolean(true), makeBoolean(true),
            makeInteger(Long.MIN_VALUE), makeInteger(Long.MIN_VALUE), makeInteger(-123456L),
            makeInteger(-123456L), makeInteger(0), makeInteger(0), makeInteger(Long.MAX_VALUE),
            makeInteger(Long.MAX_VALUE), makeString(""), makeString(""), makeString("a"),
            makeString("a"), makeString("abc"), makeString("abc"), makeFloat(-1234.5),
            makeFloat(-1234.5), makeFloat(0.0), makeFloat(0.0), makeFloat(1234.5),
            makeFloat(1234.5), makeDate(1975, 4, 9, 15, 0, 0), makeDate(1975, 4, 9, 15, 0, 0),
            makeDate(1977, 3, 9, 21, 0, 0), makeDate(1977, 3, 9, 21, 0, 0),
            makeDate(1978, 9, 11, 16, 0, 0), makeDate(1978, 9, 11, 16, 0, 0)};
    }

    /**
     * Calls {@link Value#equals(Value)} and {@link Value#equals(Object)} methods on first
     * parameter passing the second one to the called method, checks if both methods evaluate to
     * the same value and returns it.
     *
     * @param aValue instance to call <code>equals</code> method on
     * @param bValue instance to pass to <code>equals</code> method as a parameter
     * @return the result of evaluation of <code>equals</code> method
     */
    private boolean evaluateEquals(Value aValue, Value bValue) {
        boolean result = aValue.equals(bValue);
        assertEquals(aValue.equals((Object) bValue), result);
        return result;
    }

    /**
     * Tests {@link Value#equals(Value)} and {@link Value#equals(Object)} methods, comparing
     * <code>Value</code> instances holding values of the same types.
     */
    public void testEqualsSameType() {
        assertFalse(evaluateEquals(makeBoolean(false), makeBoolean(true)));
        assertTrue(evaluateEquals(makeBoolean(false), makeBoolean(false)));
        assertFalse(evaluateEquals(makeInteger(-12345L), makeInteger(12345L)));
        assertTrue(evaluateEquals(makeInteger(-12345L), makeInteger(-12345L)));
        assertFalse(evaluateEquals(makeString(""), makeString("abc")));
        assertTrue(evaluateEquals(makeString(""), makeString("")));
        assertFalse(evaluateEquals(makeFloat(-1234.5), makeFloat(1234.5)));
        assertTrue(evaluateEquals(makeFloat(-1234.5), makeFloat(-1234.5)));
    }

    /**
     * Tests {@link Value#equals(Value)} and {@link Value#equals(Object)} methods, comparing
     * <code>Value</code> instances holding values of different types.
     */
    public void testEqualsDifferentType() {
        Value[] values = generateSampleValues();
        for (int ai = 0; ai < values.length; ai++) {
            for (int bi = 0; bi < values.length; bi++) {
                Value aValue = values[ai];
                Value bValue = values[bi];
                if (aValue.getType() != bValue.getType()) {
                    assertFalse(evaluateEquals(aValue, bValue));
                    assertFalse(evaluateEquals(bValue, aValue));
                }
            }
        }
    }

    /**
     * Tests {@link Value#equals(Value)} and {@link Value#equals(Object)} methods checking
     * whether they obey requirements defined in <code>Object.equals</code>.
     */
    public void testEqualsRelationProperties() {
        Value[] values = generateSampleValues();
        for (int i = 0; i < values.length; i++) {
            // equals(null) returns false
            assertFalse(evaluateEquals(values[i], null));
        }
        for (int ai = 0; ai < values.length; ai++) {
            Value aValue = values[ai];
            for (int bi = 0; bi < values.length; bi++) {
                Value bValue = values[bi];
                // symmetric
                assertEquals(evaluateEquals(aValue, bValue), evaluateEquals(bValue, aValue));
                // reflexive
                if (aValue == bValue) {
                    assertTrue(evaluateEquals(aValue, bValue));
                }
                for (int ci = 0; ci < values.length; ci++) {
                    Value cValue = values[ci];
                    // transitive
                    if (evaluateEquals(aValue, bValue) && evaluateEquals(bValue, cValue)) {
                        assertTrue(evaluateEquals(aValue, cValue));
                    }
                }
            }
        }
    }

    /**
     * Tests {@link Value#equals(Object)} method using non-<code>Value</code> instances.
     */
    public void testEqualsObjectNotValue() {
        Value[] values = generateSampleValues();
        Object[] objects = {"", new Object(), new String[0]};
        for (int vi = 0; vi < values.length; vi++) {
            for (int oi = 0; oi < objects.length; oi++) {
                assertFalse(values[vi].equals(objects[oi]));
            }
        }
    }

    /**
     * Tests {@link Value#hashCode()} with {@link Value#equals(Value)} and
     * {@link Value#equals(Object)}, checking whether they return consistent results.
     */
    public void testHashCodeWithEquals() {
        Value[] values = generateSampleValues();
        for (int ai = 0; ai < values.length; ai++) {
            for (int bi = 0; bi < values.length; bi++) {
                Value aValue = values[ai];
                Value bValue = values[bi];
                if (evaluateEquals(aValue, bValue)) {
                    assertEquals(aValue.hashCode(), bValue.hashCode());
                }
            }
        }
    }

}
