/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.lexer;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import junit.framework.TestCase;
import pl.chyla.wyjec.lexer.Lexer;
import pl.chyla.wyjec.lexer.Token;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.TestRawString;


/**
 * Tests for {@link Lexer} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestLexer extends TestCase {

    /**
     * Lexer instance used for testing
     */
    private Lexer lexer;

    /**
     * Empty default constructor
     */
    public TestLexer() { }

    /**
     * Creates new instance of lexer
     */
    protected void setUp() throws Exception {
        super.setUp();
        lexer = new Lexer();
    }

    /**
     * Checks if the current token is an expected one, throwing assertion exception otherwise.
     *
     * @param tokenType expected type of the current token
     * @param tokenString expected string representation of the current token
     */
    private void checkToken(int tokenType, String tokenString) {
        assertEquals(lexer.getTokenType(), tokenType);
        assertEquals(lexer.getTokenString().toString(), tokenString);
    }

    /**
     * Instructs the lexer to read next token from input and then checks whether its type and
     * string representation are equal to given values.
     *
     * @param tokenType expected type of next token
     * @param tokenString expected string representation of next token
     */
    private void checkNextToken(int tokenType, String tokenString) {
        assertEquals(lexer.getNextToken(), tokenType);
        checkToken(tokenType, tokenString);
    }

    /**
     * Tests validity of state of newly created lexer (before calling
     * {@link Lexer#start(ReadableRawString)} method).
     */
    public void testBeforeStart() {
        assertEquals(lexer.getTokenStartIndex(), 0);
        assertEquals(lexer.getTokenType(), Token.TERMINATOR);
        assertEquals(lexer.getTokenString().toString(), "");
    }

    /**
     * Tests validity of lexer state immediately after calling
     * {@link Lexer#start(ReadableRawString)} method.
     */
    public void testAfterStart() {
        lexer.start(new RawString("0"));
        checkToken(Token.STRING, "0");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests validity of lexer state with empty string used as input.
     */
    public void testEmpty() {
        lexer.start(new RawString(""));
        checkToken(Token.TERMINATOR, "");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests validity of lexer state with string containing {@link Token#STRING} token used as
     * input.
     */
    public void testString() {
        lexer.start(new RawString("string"));
        checkToken(Token.STRING, "string");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests validity of lexer state with string containing {@link Token#SEPARATOR} token used
     * as input.
     */
    public void testSeparator() {
        lexer.start(new RawString("|"));
        checkToken(Token.SEPARATOR, "|");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests validity of lexer state with string containing {@link Token#FUNCTION_START} token
     * used as input.
     */
    public void testFunctionStart() {
        lexer.start(new RawString("{"));
        checkToken(Token.FUNCTION_START, "{");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests validity of lexer state with string containing {@link Token#FUNCTION_END} token
     * used as input.
     */
    public void testFunctionEnd() {
        lexer.start(new RawString("}"));
        checkToken(Token.FUNCTION_END, "}");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests validity of lexer state when getting tokens after reaching
     * {@link Token#TERMINATOR}.
     */
    public void testMultipleTerminators() {
        lexer.start(new RawString("{true}"));
        while (lexer.getNextToken() != Token.TERMINATOR) {
            // ignore non-terminator tokens
        }
        checkNextToken(Token.TERMINATOR, "");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests lexer on input string containing non-ASCII characters.
     */
    public void testNotASCII() {
        lexer.start(new RawString("{pka}w"));
        checkToken(Token.FUNCTION_START, "{");
        checkNextToken(Token.STRING, "pka");
        checkNextToken(Token.FUNCTION_END, "}");
        checkNextToken(Token.STRING, "w");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests lexer on input string consisting of many various tokens.
     */
    public void testSequence() {
        lexer.start(new RawString(" {if|{ok}|yes|no} "));
        checkToken(Token.STRING, " ");
        checkNextToken(Token.FUNCTION_START, "{");
        checkNextToken(Token.STRING, "if");
        checkNextToken(Token.SEPARATOR, "|");
        checkNextToken(Token.FUNCTION_START, "{");
        checkNextToken(Token.STRING, "ok");
        checkNextToken(Token.FUNCTION_END, "}");
        checkNextToken(Token.SEPARATOR, "|");
        checkNextToken(Token.STRING, "yes");
        checkNextToken(Token.SEPARATOR, "|");
        checkNextToken(Token.STRING, "no");
        checkNextToken(Token.FUNCTION_END, "}");
        checkNextToken(Token.STRING, " ");
        checkNextToken(Token.TERMINATOR, "");
    }

    /**
     * Tests lexer on all permutations of specific sequence of tokens (with duplicate non-string
     * tokens).
     */
    public void testSequencePermutations() {
        List tokenDescPermutations = generatePermutations(
            new TokenDescription[] {
                new TokenDescription(Token.STRING, "kotek"),
                new TokenDescription(Token.FUNCTION_START, "{"),
                new TokenDescription(Token.FUNCTION_END, "}"),
                new TokenDescription(Token.SEPARATOR, "|"),
                new TokenDescription(Token.FUNCTION_START, "{"),
                new TokenDescription(Token.FUNCTION_END, "}"),
                new TokenDescription(Token.SEPARATOR, "|")
            });
        for (Iterator iter = tokenDescPermutations.iterator(); iter.hasNext();) {
            TokenDescription[] tokenDescs = (TokenDescription[]) iter.next();
            StringBuffer expr = new StringBuffer();
            for (int i = 0; i < tokenDescs.length; i++) {
                expr.append(tokenDescs[i].getTokenString());
            }
            RawString[] exprVariants = TestRawString.buildRawStringVariants(expr.toString());
            for (int i = 0; i < exprVariants.length; i++) {
                lexer.start(exprVariants[i]);
                int exprPos = 0;
                for (int j = 0; j < tokenDescs.length; j++) {
                    TokenDescription tokenDesc = tokenDescs[j];
                    assertEquals(lexer.getTokenType(), tokenDesc.getTokenType());
                    assertEquals(lexer.getTokenString().toString(), tokenDesc.getTokenString());
                    assertEquals(lexer.getTokenStartIndex(), exprPos);
                    exprPos += tokenDesc.getTokenString().length();
                    lexer.getNextToken();
                }
            }
            assertEquals(lexer.getTokenType(), Token.TERMINATOR);
            assertEquals(lexer.getTokenStartIndex(), expr.length());
        }
    }

    /**
     * Generates a list of permutations of given array of objects. Every element of returned
     * list will be an array of objects with size equal to the size of input array, but the
     * arrays will differ with order of their elements.
     *
     * @param elements array of elements to generate permutations for
     * @return list of permutations of given array
     */
    private List generatePermutations(Object[] elements) {
        if (elements.length <= 1) {
            List permutations = new LinkedList();
            permutations.add(elements);
            return permutations;
        }
        List permutations = new LinkedList();
        Object[] subElements = new Object[elements.length - 1];
        System.arraycopy(elements, 1, subElements, 0, elements.length - 1);
        List subPermutations = generatePermutations(subElements);
        for (Iterator iter = subPermutations.iterator(); iter.hasNext();) {
            Object[] subPermutation = (Object[]) iter.next();
            for (int i = 0; i < elements.length; i++) {
                Object[] permutation = (Object[]) elements.clone();
                permutation[i] = elements[0];
                System.arraycopy(subPermutation, 0, permutation, 0, i);
                System.arraycopy(
                    subPermutation, i, permutation, i + 1, elements.length - i - 1);
                permutations.add(permutation);
            }
        }
        return permutations;
    }

    /**
     * Tests {@link Lexer#getInputString()} method on newly created {@link Lexer} instance.
     */
    public void testGetInputStringBeforeStart() {
        assertEquals(lexer.getInputString().toString(), "");
    }

    /**
     * Tests {@link Lexer#getInputString()} method after setting the lexer's input.
     */
    public void testGetInputStringAfterStart() {
        RawString rs = new RawString("{true}");
        lexer.start(rs);
        assertSame(lexer.getInputString(), rs);
    }

    /**
     * Describes single token by encapsulating its type (e.g. {@link Token#STRING}) and string
     * representation (e.g. "some token").
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private class TokenDescription {

        /**
         * Token's type
         */
        private final int tokenType;

        /**
         * Token's string representation
         */
        private final String tokenString;

        /**
         * Constructs immutable <code>TokenDescription</code> instance.
         *
         * @param tokenType token type to encapsulate
         * @param tokenString token string representation to encapsulate
         */
        TokenDescription(int tokenType, String tokenString) {
            this.tokenType = tokenType;
            this.tokenString = tokenString;
        }

        /**
         * Returns encapsulated token type
         *
         * @return token type
         */
        int getTokenType() {
            return tokenType;
        }

        /**
         * Returns encapsulated token string representation
         *
         * @return token string representation
         */
        String getTokenString() {
            return tokenString;
        }

    }

}
