/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import pl.chyla.wyjec.core.Function;


/**
 * Class describing expression tree node representing a reference to a function and its
 * parameters.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FunctionNodeDescription extends AtomNodeDescription {

    /**
     * Function referenced from the node described by this object.
     */
    private final Function function;

    /**
     * Descriptions of atom nodes used as function parameters in the node described by this
     * object.
     */
    private final AtomNodeDescription[] parameters;

    /**
     * Constructs <code>FunctionNodeDescription</code> instance, using given function and
     * descriptions of parameters as a content of described node. The actual expression tree
     * nodes for parameters will be generated from their descriptions on demand.
     *
     * @param function a function that described function node should refer to
     * @param parameters descriptions of the function parameters
     */
    public FunctionNodeDescription(Function function, AtomNodeDescription[] parameters) {
        this.function = function;
        this.parameters = parameters;
    }

    /**
     * Generates function reference node specified by this description object, together with all
     * linked parameter nodes and their (direct and indirect) references.
     *
     * @param pool a pool used to generate instances of expression tree nodes
     * @return atom node described by this object
     */
    public AtomNode makeAtomNode(ExprTreePool pool) {
        ParameterNode firstParamNode = null;
        ParameterNode lastParamNode = firstParamNode;
        for (int i = 0; i < parameters.length; i++) {
            ParameterNode paramNode = pool.getParameterNode();
            AtomNode paramAtomNode = parameters[i].makeAtomNode(pool);
            if (firstParamNode == null) {
                paramNode.setFirst(paramAtomNode);
                firstParamNode = paramNode;
            } else {
                paramNode.setAfter(lastParamNode, paramAtomNode);
            }
            lastParamNode = paramNode;
        }
        if (lastParamNode != null) {
            lastParamNode.setLast();
        }
        FunctionNode funcNode = pool.getFunctionNode();
        funcNode.set(function, firstParamNode);
        return funcNode;
    }

    /**
     * Checks in the parameters descriptions contained in this object match given sequence of
     * nodes representing function parameters.
     *
     * @param paramNodes nodes representing function's parameters to check
     * @return
     *     <code>true</code> if the parameters descriptions match parameters, <code>false</code>
     *     otherwise
     */
    private boolean isParamListMatching(ParameterNode paramNodes) {
        boolean result = true;
        ParameterNode paramNode = paramNodes;
        int paramIndex = 0;
        while (paramNode != null && paramIndex < parameters.length) {
            if (!parameters[paramIndex].isMatching(paramNode.getAtom())) {
                result = false;
                break;
            }
            paramNode = paramNode.getNext();
            paramIndex++;
        }
        if (paramNode != null || paramIndex != parameters.length) {
            result = false;
        }
        return result;
    }

    /**
     * Informs whether this description matches given real atom node. The node matches if it's
     * a function node referring to the same function as this description and it contains a
     * sequence of parameters that matches an array of parameters descriptions contained in
     * this object.
     *
     * @param atomNode atom node to check this description against.
     * @return
     *     <code>true</code> if this description matches given node, <code>false</code>
     *     otherwise
     */
    public boolean isMatching(AtomNode atomNode) {
        boolean result = true;
        if (atomNode.getAtomType() == AtomType.FUNCTION) {
            FunctionNode funcNode = atomNode.getAsFunctionNode();
            result = funcNode.getFunction() == function
                     && isParamListMatching(funcNode.getParameters());
        } else {
            result = false;
        }
        return result;
    }

}
