/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import junit.framework.TestCase;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.stdconsts.StandardConstantsParser;
import pl.chyla.wyjec.stdfuncs.StandardFunctions;
import pl.chyla.wyjec.testutils.IdEqualityWrapper;
import pl.chyla.wyjec.utils.RawString;


/**
 * Tests for {@link ExpressionTree} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestExpressionTree extends TestCase {

    /**
     * Parser instance used for testing.
     */
    private Parser parser;

    /**
     * Expression passed to the parser in the tests.
     */
    private static final String EXAMPLE_EXPRESSION = "Test: {lcurly}{+|{2}|{3}}{rcurly}";

    /**
     * Empty default constructor
     */
    public TestExpressionTree() { }

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        parser = new Parser(
            Config.getDefaultConfig(), StandardFunctions.getFunctions(),
            StandardConstantsParser.getInstance());
    }

    /**
     * Returns a list of all expression objects (objects that can be got from
     * {@link ExprTreePool}) referenced by given atom node and its sub-nodes, including the atom
     * itself.
     *
     * @param atomNode atom node to collect objects from
     * @return a list of collected objects
     */
    static List collectTreeObjects(AtomNode atomNode) {
        List objects = new LinkedList();
        objects.add(atomNode);
        if (atomNode.getAtomType() == AtomType.FUNCTION) {
            ParameterNode params = atomNode.getAsFunctionNode().getParameters();
            for (ParameterNode param = params; param != null; param = param.getNext()) {
                objects.add(param);
                objects.addAll(collectTreeObjects(param.getAtom()));
            }
        } else {
            objects.add(atomNode.getAsValueNode().getValue());
        }
        return objects;
    }

    /**
     * Returns a list of all expression objects (objects that can be got from
     * {@link ExprTreePool}) used to build given expression, including given object itself.
     *
     * @param tree expression to collect objects from
     * @return a list of collected objects
     */
    static List collectTreeObjects(ExpressionTree tree) {
        List objects = new LinkedList();
        objects.add(tree);
        objects.addAll(collectTreeObjects(tree.getTopLevelAtom()));
        return objects;
    }

    /**
     * Tests {@link ExpressionTree#makeDeepCopy(ExprTreePool)} method, checking whether objects
     * enclosed in the copies are really unique.
     */
    public void testMakeDeepCopyUniqueness() throws InvalidInputException {
        ExpressionTree baseExprTree = parser.parse(new RawString(EXAMPLE_EXPRESSION));
        List treeObjs = collectTreeObjects(baseExprTree);
        List wrappedTreeObjs = IdEqualityWrapper.wrapObjects(treeObjs);
        Set allWrappedTreeObjs = new HashSet(wrappedTreeObjs);
        ExprTreePool privPool = new ExprTreePool(1, Config.getDefaultConfig());
        for (int i = 0; i < 10; i++) {
            ExpressionTree exprTree = baseExprTree.makeDeepCopy(privPool);
            treeObjs = collectTreeObjects(exprTree);
            wrappedTreeObjs = IdEqualityWrapper.wrapObjects(treeObjs);
            Iterator iter = wrappedTreeObjs.iterator();
            while (iter.hasNext()) {
                assertFalse(allWrappedTreeObjs.contains(iter.next()));
            }
            allWrappedTreeObjs.addAll(wrappedTreeObjs);
        }
    }

    /**
     * Tests {@link ExpressionTree#makeDeepCopy(ExprTreePool)} method, checking whether a copy
     * is a semantic equivalent of its original.
     */
    public void testMakeDeepCopyEquivalence() throws InvalidInputException {
        ExpressionTree exprTree = parser.parse(new RawString(EXAMPLE_EXPRESSION));
        ExprTreePool privPool = new ExprTreePool(1, Config.getDefaultConfig());
        ExpressionTree exprTreeCopy = exprTree.makeDeepCopy(privPool);
        TestAtomNode.checkAtomNodeEquivalence(
            exprTree.getTopLevelAtom(), exprTreeCopy.getTopLevelAtom());
    }

    /**
     * Tests {@link ExpressionTree#deepRelease(ExprTreePool)} method, checking whether
     * expression tree objects actually get released to given pool.
     */
    public void testDeepRelease() throws InvalidInputException {
        ExpressionTree baseExprTree = parser.parse(new RawString(EXAMPLE_EXPRESSION));
        ExprTreePool privPool = new ExprTreePool(1, Config.getDefaultConfig());
        ExpressionTree exprTree = baseExprTree.makeDeepCopy(privPool);
        List treeObjs = TestExpressionTree.collectTreeObjects(exprTree);
        List wrappedTreeObjs = IdEqualityWrapper.wrapObjects(treeObjs);
        Set wrappedTreeObjsSet = new HashSet(wrappedTreeObjs);
        exprTree.deepRelease(privPool);
        while (true) {
            exprTree = baseExprTree.makeDeepCopy(privPool);
            treeObjs = TestExpressionTree.collectTreeObjects(exprTree);
            wrappedTreeObjs = IdEqualityWrapper.wrapObjects(treeObjs);
            if (wrappedTreeObjsSet.containsAll(wrappedTreeObjs)) {
                break;
            }
            exprTree.deepRelease(privPool);
            wrappedTreeObjsSet.addAll(wrappedTreeObjs);
        }
    }

}
