/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import junit.framework.TestCase;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.TestValue;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.repo.FunctionsRepository;
import pl.chyla.wyjec.stdconsts.StandardConstantsParser;
import pl.chyla.wyjec.stdfuncs.MergedFunctions;
import pl.chyla.wyjec.stdfuncs.StandardFunctions;
import pl.chyla.wyjec.testutils.IdEqualityWrapper;
import pl.chyla.wyjec.utils.RawString;


/**
 * Tests for {@link Parser} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestParser extends TestCase {

    /**
     * The configuration object used for parsing.
     */
    private Config config;

    /**
     * Repository of functions used by the parser.
     */
    private FunctionsRepository funcsRepo;

    /**
     * Parser instance used for testing.
     */
    private Parser parser;

    /**
     * Empty default constructor
     */
    public TestParser() { }

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        config = Config.getDefaultConfig();
        Function[] funcs = StandardFunctions.getFunctions();
        funcsRepo = new FunctionsRepository(funcs);
        parser = new Parser(config, funcs, StandardConstantsParser.getInstance());
    }

    /**
     * Returns a function from the repository with given name and accepting parameters with
     * specified types.
     *
     * @param name function's name
     * @param paramsTypes
     *     array of types of actual parameters that must be accepted by the function
     * @return the function found
     */
    private Function getFunction(String name, Type[] paramsTypes) {
        Function func = funcsRepo.lookup(new RawString(name), paramsTypes, paramsTypes.length);
        assertNotNull(func);
        return func;
    }

    /**
     * Returns string concatenation function that accepts specified number of parameters.
     *
     * @param paramsCount number of parameters to be accepted by the function
     * @return the function found
     */
    private Function getStringConcatFunction(int paramsCount) {
        Type[] paramsTypes = new Type[paramsCount];
        Arrays.fill(paramsTypes, Type.STRING);
        Function func = getFunction(config.getStringConcatFuncName(),  paramsTypes);
        assertSame(func.getResultType(), Type.STRING);
        return func;
    }

    /**
     * Returns string conversion function that accepts single parameter of given type.
     *
     * @param srcType a type of accepted parameter
     * @return the function found
     */
    private Function getStringConversionFunction(Type srcType) {
        Type[] paramsTypes = new Type[1];
        paramsTypes[0] = srcType;
        Function func = getFunction(config.getStringConvFuncName(), paramsTypes);
        assertSame(func.getResultType(), Type.STRING);
        return func;
    }

    /**
     * Checks if the parser parses given expression without errors and generates an expression
     * tree that matches given tree description.
     *
     * @param expr expression to parse
     * @param expectedTree description of expected expression tree
     */
    private void checkExpressionParsing(String expr, ExpressionTreeDescription expectedTree)
        throws InvalidInputException {
        ExpressionTree exprTree = parser.parse(new RawString(expr));
        assertTrue(expectedTree.isMatching(exprTree));
    }

    /**
     * Checks if the parser throws {@link InvalidInputException} exception during parsing given
     * expression.
     *
     * @param expr expression to parse
     */
    private void checkInvalidExpressionParsing(String expr) {
        try {
            parser.parse(new RawString(expr));
            fail();
        } catch (InvalidInputException e) {
            // OK
        }
    }

    /**
     * Tests if the parser generates proper trees for some simple expressions.
     */
    public void testSyntaxSimple() throws InvalidInputException {
        checkExpressionParsing("", new ExpressionTreeDescription(
            new ValueNodeDescription(TestValue.makeString(""))));
        checkExpressionParsing("abc", new ExpressionTreeDescription(
            new ValueNodeDescription(TestValue.makeString("abc"))));
        checkExpressionParsing("abc{5}", new ExpressionTreeDescription(
            new FunctionNodeDescription(
                getStringConcatFunction(2), new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeString("abc")),
                    new FunctionNodeDescription(
                        getStringConversionFunction(Type.INTEGER), new AtomNodeDescription[] {
                            new ValueNodeDescription(TestValue.makeInteger(5L))})})));
        checkExpressionParsing("abc{+|}def", new ExpressionTreeDescription(
            new FunctionNodeDescription(
                getStringConcatFunction(3), new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeString("abc")),
                    new FunctionNodeDescription(
                        getStringConcatFunction(1), new AtomNodeDescription[] {
                            new ValueNodeDescription(TestValue.makeString(""))}),
                    new ValueNodeDescription(TestValue.makeString("def"))})));
        checkExpressionParsing("abc{+||}", new ExpressionTreeDescription(
            new FunctionNodeDescription(
                getStringConcatFunction(2), new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeString("abc")),
                    new FunctionNodeDescription(
                        getStringConcatFunction(2), new AtomNodeDescription[] {
                            new ValueNodeDescription(TestValue.makeString("")),
                            new ValueNodeDescription(TestValue.makeString(""))})})));
        checkExpressionParsing("abc{+|||def}", new ExpressionTreeDescription(
            new FunctionNodeDescription(
                getStringConcatFunction(2), new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeString("abc")),
                    new FunctionNodeDescription(
                        getStringConcatFunction(3), new AtomNodeDescription[] {
                            new ValueNodeDescription(TestValue.makeString("")),
                            new ValueNodeDescription(TestValue.makeString("")),
                            new ValueNodeDescription(TestValue.makeString("def"))})})));
    }

    /**
     * Tests if the parser generates proper trees for nested functions.
     */
    public void testSyntaxNestedFunctions() throws InvalidInputException {
        checkExpressionParsing("{+|{+|abc|}|{+||def}}", new ExpressionTreeDescription(
            new FunctionNodeDescription(
                getStringConcatFunction(2), new AtomNodeDescription[] {
                    new FunctionNodeDescription(
                        getStringConcatFunction(2), new AtomNodeDescription[] {
                            new ValueNodeDescription(TestValue.makeString("abc")),
                            new ValueNodeDescription(TestValue.makeString(""))}),
                    new FunctionNodeDescription(
                        getStringConcatFunction(2), new AtomNodeDescription[] {
                            new ValueNodeDescription(TestValue.makeString("")),
                            new ValueNodeDescription(TestValue.makeString("def"))})})));
        int funcDepth = config.getFuncDepthSoftLimit() * 2;
        String exprStr = "b";
        for (int i = 0; i < funcDepth; i++) {
            exprStr = "{+|a|" + exprStr + "}";
        }
        AtomNodeDescription topNodeDesc = new ValueNodeDescription(TestValue.makeString("b"));
        for (int i = 0; i < funcDepth; i++) {
            AtomNodeDescription nodeDesc;
            nodeDesc = new FunctionNodeDescription(
                getStringConcatFunction(2), new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeString("a")), topNodeDesc});
            topNodeDesc = nodeDesc;
        }
        checkExpressionParsing(exprStr, new ExpressionTreeDescription(topNodeDesc));
    }

    /**
     * Tests if the parser generates proper trees for functions with the number of parameters
     * equal to the soft limit.
     */
    public void testSyntaxParamsSoftLimit() throws InvalidInputException {
        StringBuffer exprBuf = new StringBuffer();
        exprBuf.append("{+");
        for (int i = 0; i < config.getFuncActualParamsSoftLimit(); i++) {
            exprBuf.append("|");
            exprBuf.append((char) ('A' + (i % 10)));
        }
        exprBuf.append("}");
        List paramsDescs = new LinkedList();
        for (int i = 0; i < config.getFuncActualParamsSoftLimit(); i++) {
            String paramStr = Character.toString((char) ('A' + (i % 10)));
            paramsDescs.add(new ValueNodeDescription(TestValue.makeString(paramStr)));
        }
        checkExpressionParsing(exprBuf.toString(), new ExpressionTreeDescription(
            new FunctionNodeDescription(
                getStringConcatFunction(config.getFuncActualParamsSoftLimit()),
                (AtomNodeDescription[]) paramsDescs.toArray(new AtomNodeDescription[0]))));
    }

    /**
     * Tests if the parser generates proper trees for string concatenations with the number of
     * elements equal to the soft limit for the number of function parameters.
     */
    public void testConcatenationParamsSoftLimit() throws InvalidInputException {
        StringBuffer exprBuf = new StringBuffer();
        for (int i = 0; i < config.getFuncActualParamsSoftLimit(); i++) {
            exprBuf.append("{vbar}");
        }
        Function vbarFunc = getFunction("vbar", new Type[0]);
        List paramsDescs = new LinkedList();
        for (int i = 0; i < config.getFuncActualParamsSoftLimit(); i++) {
            paramsDescs.add(new FunctionNodeDescription(vbarFunc, new AtomNodeDescription[0]));
        }
        checkExpressionParsing(exprBuf.toString(), new ExpressionTreeDescription(
            new FunctionNodeDescription(
                getStringConcatFunction(config.getFuncActualParamsSoftLimit()),
                (AtomNodeDescription[]) paramsDescs.toArray(new AtomNodeDescription[0]))));
    }

    /**
     * Tests if the parser generates proper trees for constants.
     */
    public void testConstants() throws InvalidInputException {
        checkExpressionParsing("{123}", new ExpressionTreeDescription(
            new ValueNodeDescription(TestValue.makeInteger(123L))));
        checkInvalidExpressionParsing("{123|}");
        checkExpressionParsing("{123.5}", new ExpressionTreeDescription(
            new ValueNodeDescription(TestValue.makeFloat(123.5))));
        checkInvalidExpressionParsing("{123.5|}");
        checkExpressionParsing("{@1977-03-09}", new ExpressionTreeDescription(
            new ValueNodeDescription(TestValue.makeDate(1977, 3, 9, 0, 0, 0))));
        checkInvalidExpressionParsing("{@1977-03-09|}");
    }

    /**
     * Tests if the parser generates proper trees when a name matches both a function and a
     * constant.
     */
    public void testConstantAndFunction() throws InvalidInputException {
        checkExpressionParsing("{0}", new ExpressionTreeDescription(
            new ValueNodeDescription(TestValue.makeInteger(0L))));
        Function[] funcs = MergedFunctions.getFunctions(
            new Function[][] {
                StandardFunctions.getFunctions(),
                new Function[] {new FnZero()}});
        funcsRepo = new FunctionsRepository(funcs);
        parser = new Parser(config, funcs, StandardConstantsParser.getInstance());
        Function zeroFunc = getFunction("0", new Type[0]);
        checkExpressionParsing("{0}", new ExpressionTreeDescription(
            new FunctionNodeDescription(zeroFunc, new AtomNodeDescription[0])));
    }

    /**
     * Tests if the parser generates proper trees for functions with overloaded names.
     */
    public void testOverloading() throws InvalidInputException {
        Function intAddFunc = getFunction("+", new Type[] {Type.INTEGER, Type.INTEGER});
        checkExpressionParsing("{+|{5}|{6}}", new ExpressionTreeDescription(
            new FunctionNodeDescription(
                intAddFunc, new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeInteger(5L)),
                    new ValueNodeDescription(TestValue.makeInteger(6L))})));
        Function strAddFunc = getFunction("+", new Type[] {Type.STRING, Type.STRING});
        checkExpressionParsing("{+|ab|cd}", new ExpressionTreeDescription(
            new FunctionNodeDescription(
                strAddFunc, new AtomNodeDescription[] {
                    new ValueNodeDescription(TestValue.makeString("ab")),
                    new ValueNodeDescription(TestValue.makeString("cd"))})));
        checkInvalidExpressionParsing("{+|{4}|abc}");
    }

    /**
     * Tests if the parser properly handles non-existing functions.
     */
    public void testInvalidFunction() {
        checkInvalidExpressionParsing("{xyz}");
        char aFnName = (char) ('+' | 0x80);
        checkInvalidExpressionParsing("{" + aFnName + "|abc}");
        char bFnName = (char) ('+' | 0x100);
        checkInvalidExpressionParsing("{" + bFnName + "|abc}");
    }

    /**
     * Tests if the parser properly handles functions with invalid parameters passed.
     */
    public void testInvalidFunctionParameters() {
        checkInvalidExpressionParsing("{str|{4}|{5}}");
        checkInvalidExpressionParsing("{str|{4}|x}");
        checkInvalidExpressionParsing("{+}");
        checkInvalidExpressionParsing("{+|{4}|x}");
    }

    /**
     * Tests if the parser properly handles functions with invalid type of one of optional
     * parameters.
     */
    public void testInvalidFunctionOptionalParameters() {
        StringBuffer buf = new StringBuffer();
        buf.append("{+");
        for (int i = 0; i < Config.FUNC_REQ_PARAMS_COUNT_LIMIT + 1; i++) {
            buf.append("|ss");
        }
        buf.append("|{1}}");
        checkInvalidExpressionParsing(buf.toString());
        buf.setLength(0);
        buf.append("{+");
        for (int i = 0; i < Config.FUNC_REQ_PARAMS_COUNT_LIMIT + 2; i++) {
            buf.append("|{" + i + "}");
        }
        buf.append("|ss}");
        checkInvalidExpressionParsing(buf.toString());
    }

    /**
     * Tests if the parser properly handles implicit conversion in case no matching conversion
     * function is available.
     */
    public void testMissingConversionFunction() {
        config.setStringConvFuncName("___");
        Function[] funcs = StandardFunctions.getFunctions();
        funcsRepo = new FunctionsRepository(funcs);
        parser = new Parser(config, funcs, StandardConstantsParser.getInstance());
        checkInvalidExpressionParsing("Value: {123}");
        checkInvalidExpressionParsing("Value: {123.5}");
        checkInvalidExpressionParsing("Value: {@1977-03-09}");
    }

    /**
     * Tests if the parser properly handles implicit conversion in case a function configured as
     * conversion function has invalid return type.
     */
    public void testInvalidConversionFunction() {
        config.setStringConvFuncName("+");
        Function[] funcs = StandardFunctions.getFunctions();
        funcsRepo = new FunctionsRepository(funcs);
        parser = new Parser(config, funcs, StandardConstantsParser.getInstance());
        checkInvalidExpressionParsing("Value: {123}");
        checkInvalidExpressionParsing("Value: {123.5}");
        checkInvalidExpressionParsing("Value: {@1977-03-09}");
    }

    /**
     * Tests if the parser properly handles implicit concatenation in case no matching
     * concatenation function is available.
     */
    public void testMissingConcatenationFunction() {
        config.setStringConcatFuncName("___");
        Function[] funcs = StandardFunctions.getFunctions();
        funcsRepo = new FunctionsRepository(funcs);
        parser = new Parser(config, funcs, StandardConstantsParser.getInstance());
        checkInvalidExpressionParsing("Value: {upper|xyz}");
    }

    /**
     * Tests if the parser properly handles implicit concatenation in case a function configured
     * as concatenation function has invalid return type.
     */
    public void testInvalidConcatenationFunction() {
        config.setStringConcatFuncName("=");
        Function[] funcs = StandardFunctions.getFunctions();
        funcsRepo = new FunctionsRepository(funcs);
        parser = new Parser(config, funcs, StandardConstantsParser.getInstance());
        checkInvalidExpressionParsing("Value: {upper|xyz}");
    }

    /**
     * Tests if the parser properly handles expressions with syntax errors.
     */
    public void testSyntaxInvalid() {
        String[] invalidExpressions = {
            "abc{", "abc|", "abc}", "{", "|", "}", "{abc", "|abc", "}abc", "{}", "{|}", "}{",
            "abc{}", "abc{|}", "abc{|5}", "abc{+def}", "abc{+{+|}}}", "abc{+|{+|}",
            "abc{+|{+|}abc", "abc{5}{", "abc{5}|", "abc{5}}"};
        for (int i = 0; i < invalidExpressions.length; i++) {
            checkInvalidExpressionParsing(invalidExpressions[i]);
        }
    }

    /**
     * Tests if the parser provides valid detailed information about an error inside
     * <code>InvalidInputException</code> instance.
     */
    public void testSyntaxInvalidWithDetails() {
        String expression = "{+|abcdef|{4.}}";
        try {
            parser.parse(new RawString(expression));
            fail();
        } catch (InvalidInputException e) {
            assertEquals(e.getInputString(), expression);
            assertEquals(e.getErrorPos(), 11);
        }
        expression = "{+|abc|{4}";
        try {
            parser.parse(new RawString(expression));
            fail();
        } catch (InvalidInputException e) {
            assertEquals(e.getInputString(), expression);
            assertEquals(e.getErrorPos(), 10);
        }
    }

    /**
     * Tests if the parser's constructor properly notifies about invalid parameters.
     */
    public void testInvalidParserParameters() {
        Function[] funcs = StandardFunctions.getFunctions();
        try {
            funcsRepo = new FunctionsRepository(funcs);
            parser = new Parser(null, funcs, StandardConstantsParser.getInstance());
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
        try {
            funcsRepo = new FunctionsRepository(funcs);
            parser = new Parser(config, null, StandardConstantsParser.getInstance());
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
        try {
            funcsRepo = new FunctionsRepository(funcs);
            parser = new Parser(config, funcs, null);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
    }

    /**
     * Tests the parser using {@link Config#getZeroMemLimitsConfig()} configuration.
     */
    public void testZeroMemLimitsConfig() throws InvalidInputException {
        config = Config.getZeroMemLimitsConfig();
        Function[] funcs = StandardFunctions.getFunctions();
        funcsRepo = new FunctionsRepository(funcs);
        parser = new Parser(config, funcs, StandardConstantsParser.getInstance());
        checkExpressionParsing("{+|{+|abc|}|{+||def}}", new ExpressionTreeDescription(
            new FunctionNodeDescription(
                getStringConcatFunction(2), new AtomNodeDescription[] {
                    new FunctionNodeDescription(
                        getStringConcatFunction(2), new AtomNodeDescription[] {
                            new ValueNodeDescription(TestValue.makeString("abc")),
                            new ValueNodeDescription(TestValue.makeString(""))}),
                    new FunctionNodeDescription(
                        getStringConcatFunction(2), new AtomNodeDescription[] {
                            new ValueNodeDescription(TestValue.makeString("")),
                            new ValueNodeDescription(TestValue.makeString("def"))})})));
    }

    /**
     * Tests whether the parser reuses tree objects in consecutive parsing operations.
     */
    public void testReusingTreeObjects() throws InvalidInputException {
        RawString exprStr = new RawString("Test: {upper|sum: {+|{2}|{3}}}");
        ExpressionTree exprTree = parser.parse(exprStr);
        List treeObjs = TestExpressionTree.collectTreeObjects(exprTree);
        List wrappedTreeObjs = IdEqualityWrapper.wrapObjects(treeObjs);
        Set wrappedTreeObjsSet = new HashSet(wrappedTreeObjs);
        while (true) {
            exprTree = parser.parse(exprStr);
            treeObjs = TestExpressionTree.collectTreeObjects(exprTree);
            wrappedTreeObjs = IdEqualityWrapper.wrapObjects(treeObjs);
            if (wrappedTreeObjsSet.containsAll(wrappedTreeObjs)) {
                break;
            }
            wrappedTreeObjsSet.addAll(wrappedTreeObjs);
        }
    }

}
