/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.repo;

import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.OptParameters;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.core.functions.IntegerFunction;
import pl.chyla.wyjec.core.optparameters.BooleanOptParameters;
import pl.chyla.wyjec.core.optparameters.DateOptParameters;
import pl.chyla.wyjec.core.optparameters.FloatOptParameters;
import pl.chyla.wyjec.core.optparameters.IntegerOptParameters;
import pl.chyla.wyjec.core.optparameters.StringOptParameters;
import pl.chyla.wyjec.core.parameters.BooleanParameter;
import pl.chyla.wyjec.core.parameters.DateParameter;
import pl.chyla.wyjec.core.parameters.FloatParameter;
import pl.chyla.wyjec.core.parameters.IntegerParameter;
import pl.chyla.wyjec.core.parameters.StringParameter;


/**
 * Generic implementations of a function for testing purposes. It may be used to implement
 * functions with different names and signatures, depending on the constructor's parameters.
 * The result of evaluation of the function is constant integer value (0).
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
final class FnGeneric extends IntegerFunction {

    /**
     * Function's name.
     */
    final String name;

    /**
     * Function's signature.
     */
    final FunctionSignature signature;

    /**
     * Constructs <code>FnGeneric</code> using given name no parameters for the function.
     *
     * @param name function's name
     */
    FnGeneric(String name) {
        this(name, new Type[0], Type.NONE);
    }

    /**
     * Constructs <code>FnGeneric</code> using given name and types of parameters for the
     * function.
     *
     * @param name function's name
     * @param reqParamsTypes array of types of required parameters
     * @param optParamsType
     *     type of optional parameters or {@link Type#NONE} if the function doesn't accept
     *     optional parameters
     */
    FnGeneric(String name, Type[] reqParamsTypes, Type optParamsType) {
        this.name = name;
        Parameter[] params = new Parameter[reqParamsTypes.length];
        for (int i = 0; i < reqParamsTypes.length; i++) {
            params[i] = getParameterForType(reqParamsTypes[i]);
        }
        OptParameters optParams = getOptParametersForType(optParamsType);
        signature = new FunctionSignature(params, optParams);
    }

    /**
     * {@inheritDoc}
     *
     * @returns {@inheritDoc}
     */
    public String getName() {
        return name;
    }

    /**
     * {@inheritDoc}
     *
     * @returns {@inheritDoc}
     */
    public FunctionSignature getSignature() {
        return signature;
    }

    /**
     * {@inheritDoc}
     *
     * @return constant value <code>0</code>
     */
    protected long execute() {
        return 0;
    }

    /**
     * Creates instance of one of subclasses of {@link Parameter} that describes required
     * parameter of specified type.
     *
     * @param paramType type of required parameter of a function
     * @return new {@link Parameter} instance that corresponds to given type
     */
    private Parameter getParameterForType(Type paramType) {
        Parameter param;
        switch (paramType.getEncodedType()) {
        case Type.BOOLEAN_ENCODED:
            param = new BooleanParameter();
            break;
        case Type.INTEGER_ENCODED:
            param = new IntegerParameter();
            break;
        case Type.STRING_ENCODED:
            param = new StringParameter();
            break;
        case Type.FLOAT_ENCODED:
            param = new FloatParameter();
            break;
        case Type.DATE_ENCODED:
            param = new DateParameter();
            break;
        default:
            throw new IllegalArgumentException(
                "Invalid parameter type: " + paramType.getName());
        }
        return param;
    }

    /**
     * Creates instance of one of subclasses of {@link OptParameters} that describes optional
     * parameters of specified type.
     *
     * @param optParamsType type of optional parameters of a function
     * @return new {@link OptsParameters} instance that corresponds to given type
     */
    private OptParameters getOptParametersForType(Type optParamsType) {
        OptParameters optParams;
        switch (optParamsType.getEncodedType()) {
        case Type.BOOLEAN_ENCODED:
            optParams = new BooleanOptParameters();
            break;
        case Type.INTEGER_ENCODED:
            optParams = new IntegerOptParameters();
            break;
        case Type.STRING_ENCODED:
            optParams = new StringOptParameters();
            break;
        case Type.FLOAT_ENCODED:
            optParams = new FloatOptParameters();
            break;
        case Type.DATE_ENCODED:
            optParams = new DateOptParameters();
            break;
        case Type.NONE_ENCODED:
            optParams = null;
            break;
        default:
            throw new IllegalArgumentException(
                "Invalid optional parameters type: " + optParamsType.getName());
        }
        return optParams;
    }

}
