/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.repo;

import java.util.Arrays;
import junit.framework.TestCase;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.stdfuncs.StandardFunctions;
import pl.chyla.wyjec.utils.ArrayUtils;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;
import pl.chyla.wyjec.utils.TestRawString;


/**
 * Tests for {@link FunctionsRepository} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestFunctionsRepository extends TestCase {

    /**
     * Empty default constructor
     */
    public TestFunctionsRepository() { }

    /**
     * Calls {@link FunctionsRepository#lookup(ReadableRawString, Type[], int)} on given
     * repository, searching for method with given name that accepts actual parameters with
     * specified types, and checks if the result is the same as given expected result.
     *
     * @param repo repository to use for lookup
     * @param funcName function's name
     * @param paramsTypes types of actual parameters
     * @param expFunc expected value returned by lookup method, may be <code>null</code>
     */
    private void checkLookup(
        FunctionsRepository repo, String funcName, Type[] paramsTypes, Function expFunc) {
        RawString[] funcNameVariants = TestRawString.buildRawStringVariants(funcName);
        Type[] paramsTypesX = new Type[paramsTypes.length * 2];
        Arrays.fill(paramsTypesX, Type.NONE);
        ArrayUtils.copyArray(paramsTypes, paramsTypesX);
        for (int i = 0; i < funcNameVariants.length; i++) {
            assertSame(
                repo.lookup(funcNameVariants[i], paramsTypes, paramsTypes.length), expFunc);
            assertSame(
                repo.lookup(funcNameVariants[i], paramsTypesX, paramsTypes.length), expFunc);
        }
    }

    /**
     * Checks if the {@link FunctionsRepository#FunctionsRepository(Function[])} constructor
     * refuses given sequence of functions for some reason by throwing
     * <code>IllegalArgumentException</code>.
     *
     * @param functions array of functions to pass to the repository
     */
    private void checkInvalidFunctionsSet(Function[] functions) {
        try {
            new FunctionsRepository(functions);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
        Function[] revFunctions = new Function[functions.length];
        for (int i = 0; i < functions.length; i++) {
            revFunctions[revFunctions.length - 1 - i] = functions[i];
        }
        try {
            new FunctionsRepository(revFunctions);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
    }

    /**
     * Checks if the repository refuses function with given name (by throwing
     * <code>IllegalArgumentException</code>).
     *
     * @param name function name to check
     */
    private void checkInvalidFunctionName(String name) {
        checkInvalidFunctionsSet(new Function[] {new FnGeneric(name)});
        checkInvalidFunctionsSet(new Function[] {new FnGeneric("ok"), new FnGeneric(name)});
    }

    /**
     * Tests construction of the repository without any functions given.
     */
    public void testRepositoryWithoutFunctions() {
        FunctionsRepository repo = new FunctionsRepository(new Function[0]);
        checkLookup(repo, "+", new Type[0], null);
        checkLookup(repo, "+", new Type[] {Type.STRING}, null);
    }

    /**
     * Tests construction of the repository using functions with invalid names.
     */
    public void testInvalidFunctionName() {
        String[] invalidNames = {"", "abc\u0080", "a\u0100bc"};
        for (int i = 0; i < invalidNames.length; i++) {
            checkInvalidFunctionName(invalidNames[i]);
        }
        StringBuffer nameBuf = new StringBuffer();
        for (int i = 0; i <= Config.FUNC_NAME_LENGTH_LIMIT; i++) {
            nameBuf.append("x");
        }
        checkInvalidFunctionName(nameBuf.toString());
    }

    /**
     * Tests {@link FunctionsRepository#lookup(ReadableRawString, Type[], int)} with empty
     * function name.
     */
    public void testLookupEmptyName() {
        FunctionsRepository repo = new FunctionsRepository(StandardFunctions.getFunctions());
        checkLookup(repo, "", new Type[0], null);
        checkLookup(repo, "", new Type[] {Type.STRING}, null);
    }

    /**
     * Tests {@link FunctionsRepository#lookup(ReadableRawString, Type[], int)} with
     * parameter-less functions.
     */
    public void testLookupNoParamsFuncs() {
        Function fnF1 = new FnGeneric("f1");
        Function fnF2 = new FnGeneric("f2");
        Function fnAbc = new FnGeneric("abc");
        Function fn1 = new FnGeneric("1");
        Function[] funcs = {fnF1, fnF2, fnAbc, fn1};
        FunctionsRepository repo = new FunctionsRepository(funcs);
        checkLookup(repo, "f1", new Type[0], fnF1);
        checkLookup(repo, "f2", new Type[0], fnF2);
        checkLookup(repo, "abc", new Type[0], fnAbc);
        checkLookup(repo, "1", new Type[0], fn1);
        checkLookup(repo, "f1", new Type[] {Type.STRING}, null);
        checkLookup(repo, "f2", new Type[] {Type.INTEGER}, null);
        checkLookup(repo, "abc", new Type[] {Type.FLOAT}, null);
        checkLookup(repo, "1", new Type[] {Type.DATE}, null);
        checkLookup(repo, "f3", new Type[0], null);
        checkLookup(repo, "2", new Type[0], null);
    }

    /**
     * Tests {@link FunctionsRepository#lookup(ReadableRawString, Type[], int)} with simple
     * functions.
     */
    public void testLookupSimpleFuncs() {
        Function fnAbc = new FnGeneric(
            "abc", new Type[] {Type.STRING, Type.INTEGER}, Type.NONE);
        Function fnDef = new FnGeneric(
            "def", new Type[] {Type.INTEGER, Type.INTEGER, Type.DATE}, Type.FLOAT);
        Function[] funcs = {fnAbc, fnDef};
        FunctionsRepository repo = new FunctionsRepository(funcs);
        checkLookup(repo, "abc", new Type[0], null);
        checkLookup(repo, "abc", new Type[] {Type.STRING}, null);
        checkLookup(repo, "abc", new Type[] {Type.STRING, Type.INTEGER, Type.DATE}, null);
        checkLookup(repo, "abc", new Type[] {Type.STRING, Type.INTEGER}, fnAbc);
        checkLookup(repo, "abc", new Type[] {Type.INTEGER, Type.INTEGER, Type.DATE}, null);
        checkLookup(repo, "def", new Type[0], null);
        checkLookup(repo, "def", new Type[] {Type.INTEGER}, null);
        checkLookup(repo, "def", new Type[] {Type.INTEGER, Type.INTEGER}, null);
        checkLookup(repo, "def", new Type[] {Type.INTEGER, Type.INTEGER, Type.INTEGER}, null);
        checkLookup(repo, "def", new Type[] {Type.INTEGER, Type.INTEGER, Type.DATE}, fnDef);
        checkLookup(
            repo, "def", new Type[] {Type.INTEGER, Type.INTEGER, Type.DATE, Type.FLOAT}, fnDef);
        checkLookup(
            repo, "def",
            new Type[] {Type.INTEGER, Type.INTEGER, Type.DATE, Type.FLOAT, Type.FLOAT}, fnDef);
        checkLookup(
            repo, "def",
            new Type[] {Type.INTEGER, Type.INTEGER, Type.DATE, Type.FLOAT, Type.DATE}, null);
    }

    /**
     * Tests {@link FunctionsRepository#lookup(ReadableRawString, Type[], int)} with very long
     * sequences of actual parameters.
     */
    public void testLookupManyArgs() {
        Function fnAbc = new FnGeneric("abc", new Type[] {Type.STRING}, Type.INTEGER);
        Function[] funcs = {fnAbc};
        FunctionsRepository repo = new FunctionsRepository(funcs);
        checkLookup(repo, "abc", new Type[] {Type.STRING}, fnAbc);
        checkLookup(repo, "abc", new Type[] {Type.STRING, Type.INTEGER}, fnAbc);
        Type[] paramsTypes = new Type[Config.FUNC_REQ_PARAMS_COUNT_LIMIT * 100];
        Arrays.fill(paramsTypes, Type.INTEGER);
        paramsTypes[0] = Type.STRING;
        checkLookup(repo, "abc", paramsTypes, fnAbc);
        Arrays.fill(paramsTypes, Type.INTEGER);
        paramsTypes[0] = Type.STRING;
        paramsTypes[1] = Type.STRING;
        checkLookup(repo, "abc", paramsTypes, null);
        Arrays.fill(paramsTypes, Type.INTEGER);
        paramsTypes[0] = Type.STRING;
        paramsTypes[paramsTypes.length - 2] = Type.STRING;
        checkLookup(repo, "abc", paramsTypes, null);
    }

    /**
     * Tests construction of the repository with more than
     * {@link Config#FUNC_REQ_PARAMS_COUNT_LIMIT} parameters.
     */
    public void testTooManyReqParams() {
        Type[] reqParamsTypes = new Type[Config.FUNC_REQ_PARAMS_COUNT_LIMIT + 1];
        Arrays.fill(reqParamsTypes, Type.INTEGER);
        Function fnAbc = new FnGeneric("abc", reqParamsTypes, Type.STRING);
        checkInvalidFunctionsSet(new Function[] {fnAbc});
    }

    /**
     * Tests construction of the repository using overloaded functions with non-disjuctive
     * parameters declarations.
     */
    public void testInvalidOverloading() {
        Function fnAbc1 = new FnGeneric("abc", new Type[] {Type.STRING}, Type.INTEGER);
        Function fnAbc2 = new FnGeneric("abc", new Type[] {Type.STRING}, Type.INTEGER);
        Function fnAbc3 = new FnGeneric(
            "abc", new Type[] {Type.STRING, Type.INTEGER}, Type.INTEGER);
        Function fnAbc4 = new FnGeneric(
            "abc", new Type[] {Type.STRING, Type.STRING}, Type.INTEGER);
        checkInvalidFunctionsSet(new Function[] {fnAbc1, fnAbc2});
        checkInvalidFunctionsSet(new Function[] {fnAbc1, fnAbc3});
        checkInvalidFunctionsSet(new Function[] {fnAbc1, fnAbc3, fnAbc4});
        Function fnDef1 = new FnGeneric("def", new Type[] {Type.STRING}, Type.INTEGER);
        Function fnDef2 = new FnGeneric("def", new Type[] {Type.STRING}, Type.DATE);
        checkInvalidFunctionsSet(new Function[] {fnDef1, fnDef2});
        Function fnGhi1 = new FnGeneric("ghi", new Type[] {}, Type.INTEGER);
        Function fnGhi2 = new FnGeneric("ghi", new Type[] {}, Type.STRING);
        Function fnGhi3 = new FnGeneric("ghi", new Type[] {}, Type.DATE);
        checkInvalidFunctionsSet(new Function[] {fnGhi1, fnGhi2, fnGhi3});
    }

    /**
     * Tests construction of the repository using overloaded functions and
     * {@link FunctionsRepository#lookup(ReadableRawString, Type[], int)} method on the
     * constructed repository.
     */
    public void testOverloading() {
        Function fnAbc1 = new FnGeneric("abc", new Type[] {Type.INTEGER}, Type.INTEGER);
        Function fnAbc2 = new FnGeneric("abc", new Type[] {Type.STRING}, Type.STRING);
        Function fnAbc3 = new FnGeneric("abc", new Type[] {Type.DATE}, Type.DATE);
        Function[] funcs = {fnAbc1, fnAbc2, fnAbc3};
        FunctionsRepository repo = new FunctionsRepository(funcs);
        checkLookup(repo, "abc", new Type[] {Type.INTEGER}, fnAbc1);
        checkLookup(repo, "abc", new Type[] {Type.STRING}, fnAbc2);
        checkLookup(repo, "abc", new Type[] {Type.DATE}, fnAbc3);
        checkLookup(repo, "abc", new Type[] {Type.BOOLEAN}, null);
        Function fnDef1 = new FnGeneric(
            "def", new Type[] {Type.INTEGER, Type.STRING}, Type.INTEGER);
        Function fnDef2 = new FnGeneric(
            "def", new Type[] {Type.INTEGER, Type.STRING, Type.STRING}, Type.NONE);
        Function fnDef3 = new FnGeneric(
            "def", new Type[] {Type.INTEGER, Type.STRING, Type.STRING, Type.INTEGER},
            Type.NONE);
        Function fnDef4 = new FnGeneric("def", new Type[] {Type.INTEGER}, Type.INTEGER);
        funcs = new Function[] {fnDef1, fnDef2, fnDef3, fnDef4};
        repo = new FunctionsRepository(funcs);
        checkLookup(repo, "def", new Type[] {Type.INTEGER, Type.STRING}, fnDef1);
        checkLookup(
            repo, "def", new Type[] {Type.INTEGER, Type.STRING, Type.INTEGER}, fnDef1);
        checkLookup(
            repo, "def", new Type[] {Type.INTEGER, Type.STRING, Type.STRING}, fnDef2);
        checkLookup(
            repo, "def", new Type[] {Type.INTEGER, Type.STRING, Type.STRING, Type.STRING},
            null);
        checkLookup(
            repo, "def", new Type[] {Type.INTEGER, Type.STRING, Type.STRING, Type.INTEGER},
            fnDef3);
        checkLookup(
            repo, "def", new Type[] {Type.INTEGER, Type.STRING, Type.INTEGER, Type.INTEGER},
            fnDef1);
        checkLookup(
            repo, "def", new Type[] {Type.INTEGER, Type.INTEGER, Type.INTEGER, Type.INTEGER},
            fnDef4);
        checkLookup(
            repo, "def", new Type[] {Type.INTEGER, Type.INTEGER, Type.INTEGER, Type.STRING},
            null);
    }

}
