/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.stdconsts;

import java.sql.Date;
import java.text.SimpleDateFormat;
import java.util.Locale;
import java.util.TimeZone;
import junit.framework.TestCase;
import pl.chyla.wyjec.core.TestValue;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;
import pl.chyla.wyjec.utils.TestRawString;


/**
 * Tests for {@link pl.chyla.wyjec.stdconsts.DateConstantsParser} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestDateConstantsParser extends TestCase {

    /**
     * Number of milliseconds per day.
     */
    private static final long MILLISECONDS_PER_DAY = 1000L * 60L * 60L * 24L;

    /**
     * Instance of constants parser to use by tests.
     */
    private DateConstantsParser constsParser;

    /**
     * Empty default constructor
     */
    public TestDateConstantsParser() { }

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        constsParser = DateConstantsParser.getInstance();
    }

    /**
     * Tests {@link DateConstantsParser#parseConstant(Value, ReadableRawString)} method,
     * passing invalid strings to parse.
     */
    public void testParseConstantInvalid() {
        String[] constStrings = {
            "", " ", "   ", "@", "1977-03-09", "1977-03-09 21:00:00",
            "@    -  -  ",  "@x977-03-09", "@1x77-03-09", "@19x7-03-09", "@197x-03-09",
            "@1977-x3-09", "@1977-0x-09", "@1977-03-x9", "@1977-03-0x",
            "@    -  -     :  :  ", "@1977 03-09 21:00:00", "@1977-03 09 21:00:00",
            "@1977-03-09,21:00:00", "@1977-03-09 21 00:00", "@1977-03-09 21:00 00",
            "@x977-03-09 21:00:00", "@1x77-03-09 21:00:00", "@19x7-03-09 21:00:00",
            "@197x-03-09 21:00:00", "@1977-x3-09 21:00:00", "@1977-0x-09 21:00:00",
            "@1977-03-x9 21:00:00", "@1977-03-0x 21:00:00", "@1977-03-09 x1:00:00",
            "@1977-03-09 2x:00:00", "@1977-03-09 21:x0:00", "@1977-03-09 21:0x:00",
            "@1977-03-09 21:00:x0", "@1977-03-09 21:00:0x", "@1977-03-09 21:00:00 ",
            " @1977-03-09 21:00:00"};
        RawString[] constRStrings = TestRawString.buildRawStringsVariants(constStrings);
        Value value = new Value(0);
        for (int i = 0; i < constRStrings.length; i++) {
            assertFalse(constsParser.parseConstant(value, constRStrings[i]));
        }
    }

    /**
     * Tests {@link DateConstantsParser#parseConstant(Value, ReadableRawString)} method, using
     * short form to represent a date (@yyyy-mm-dd).
     */
    public void testParseConstantValidShort() {
        long[] datesMillis = {
            0L * MILLISECONDS_PER_DAY, 10L * MILLISECONDS_PER_DAY, 100L * MILLISECONDS_PER_DAY,
            1000L * MILLISECONDS_PER_DAY, 10000L * MILLISECONDS_PER_DAY,
            100000L * MILLISECONDS_PER_DAY};
        SimpleDateFormat dayFormat = new SimpleDateFormat("yyyy-MM-dd", Locale.ENGLISH);
        Value value = new Value(0);
        dayFormat.setTimeZone(TimeZone.getTimeZone("GMT"));
        for (int i = 0; i < datesMillis.length; i++) {
            Date d = new Date(datesMillis[i]);
            String dateStr = "@" + dayFormat.format(d);
            RawString[] dateRStrings = TestRawString.buildRawStringVariants(dateStr);
            for (int rsi = 0; rsi < dateRStrings.length; rsi++) {
                assertTrue(constsParser.parseConstant(value, dateRStrings[rsi]));
                assertEquals(value, TestValue.makeDate(datesMillis[i]));
            }
        }
        RawString[] dateRStrings = TestRawString.buildRawStringVariants("@1977-03-09");
        for (int rsi = 0; rsi < dateRStrings.length; rsi++) {
            assertTrue(constsParser.parseConstant(value, dateRStrings[rsi]));
            assertEquals(value, TestValue.makeDate(226713600000L));
        }
    }

    /**
     * Tests {@link DateConstantsParser#parseConstant(Value, ReadableRawString)} method, using
     * short form to represent a date (@yyyy-mm-dd HH:MM::SS).
     */
    public void testParseConstantValidLong() {
        long[] datesMillis = {
            0 * MILLISECONDS_PER_DAY, 10 * MILLISECONDS_PER_DAY, 100 * MILLISECONDS_PER_DAY,
            1000 * MILLISECONDS_PER_DAY, 10000 * MILLISECONDS_PER_DAY,
            100000 * MILLISECONDS_PER_DAY};
        long[] timesMillis = {
            0L, 1L * 1000L, 10L * 1000L, 100L * 1000L, 1000L * 1000L, 10000L * 1000L,
            50000L * 1000L};
        SimpleDateFormat dayFormat = new SimpleDateFormat(
            "yyyy-MM-dd HH:mm:ss", Locale.ENGLISH);
        Value value = new Value(0);
        dayFormat.setTimeZone(TimeZone.getTimeZone("GMT"));
        for (int i = 0; i < datesMillis.length; i++) {
            for (int j = 0; j < timesMillis.length; j++) {
                Date d = new Date(datesMillis[i] + timesMillis[j]);
                String dateStr = "@" + dayFormat.format(d);
                RawString[] dateRStrings = TestRawString.buildRawStringVariants(dateStr);
                for (int rsi = 0; rsi < dateRStrings.length; rsi++) {
                    assertTrue(constsParser.parseConstant(value, dateRStrings[rsi]));
                    assertEquals(value, TestValue.makeDate(datesMillis[i] + timesMillis[j]));
                }
            }
        }
    }

}
