/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.stdconsts;

import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.TestRawString;
import junit.framework.TestCase;


/**
 * Tests for {@link pl.chyla.wyjec.stdconsts.FloatConstantsParser} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestFloatConstantsParser extends TestCase {

    /**
     * Sequences of digits to be used in float constants as integer, fractional or exponent
     * part.
     */
    private static final String[] digitsStrings = {
        "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "1234567890", "9876543210",
        "12345678901234567890123456789", "987654321098765432109876543210",
        "99", "9999999999", "999999999999999999999999999999",
        "000000000000000000001234567890", "000000000000000000009876543210"};

    /**
     * Strings to be used in float constants as sign before integer digits or exponent digits.
     */
    private static final String[] signStrings = {"", "+", "-"};

    /**
     * Instance of constants parser to use by tests.
     */
    private FloatConstantsParser constsParser;

    /**
     * Empty default constructor
     */
    public TestFloatConstantsParser() { }

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        constsParser = FloatConstantsParser.getInstance();
    }

    /**
     * Tests {@link FloatConstantsParser#parseConstant(Value, ReadableRawString)} method,
     * passing invalid strings to parse.
     */
    public void testParseConstantInvalid() {
        String[] constStrings = {
            "", " ", "   ", "0", "12345", "134.", ".123", ".123e0", "123e", "e2", "123.e2",
            " 0", "0 ", "1.f", "1ef", "f.2", "fe2",
            "f1.234", "f1234e2", "f1.234e2", "1.234f", "1234e2f", "1.234ef"};
        RawString[] constRStrings = TestRawString.buildRawStringsVariants(constStrings);
        Value value = new Value(0);
        for (int i = 0; i < constRStrings.length; i++) {
            assertFalse(constsParser.parseConstant(value, constRStrings[i]));
        }
    }

    /**
     * Evaluates signum function for given number, i.e.:
     * <ul>
     * <li>0.0 for 0.0 (preserving its sign)</li>
     * <li>1.0 for positive numbers (including positive infinity)</li>
     * <li>-1.0 for negative numbers (including negative infinity)</li>
     * <li>NaN for NaN</li>
     * </ul>
     *
     * @param value value to calculate signum for
     * @return signum of given value
     */
    private double signum(double value) {
        double result;
        if (Double.isNaN(value)) {
            result = Double.NaN;
        } else if (value == 0.0) {
            result = value;
        } else if (value > 0.0) {
            result = 1.0;
        } else {
            result = -1.0;
        }
        return result;
    }

    /**
     * Verifies whether value returned by parser is equal to given expected one. Small
     * differences between finite values will be accepted, because many double values can't be
     * written precisely using decimal representation (additional differences may be caused by
     * different algorithms used for parsing string representations of numbers).
     *
     * @param expectedValue expected value of parsed number
     * @param parsedValue the actual value returned by parser
     */
    private void checkParsedValue(double expectedValue, double parsedValue) {
        assertFalse(Double.isNaN(expectedValue));
        assertFalse(Double.isNaN(parsedValue));
        assertEquals(signum(expectedValue), signum(parsedValue), 0.0);
        if (Double.isInfinite(expectedValue)) {
            assertTrue(Double.isInfinite(parsedValue));
        } else {
            assertFalse(Double.isInfinite(parsedValue));
            if (Math.abs(expectedValue) < Double.MIN_VALUE) {
                assertTrue(Math.abs(parsedValue) < Double.MIN_VALUE);
            } else {
                assertEquals(1.0, parsedValue / expectedValue, 0.000000000000001);
            }
        }
    }

    /**
     * Tests {@link FloatConstantsParser#parseConstant(Value, ReadableRawString)} method with
     * simple form of float constants: [<sign>]<digits>.<digits>.
     */
    public void testParseConstantSimpleForm() {
        Value value = new Value(0);
        for (int si = 0; si < signStrings.length; si++) {
            for (int ii = 0; ii < digitsStrings.length; ii++) {
                for (int fi = 0; fi < digitsStrings.length; fi++) {
                    String constStr = signStrings[si] + digitsStrings[ii]
                                      + "." + digitsStrings[fi];
                    double expectedValue = Double.parseDouble(constStr);
                    RawString[] constRStrings = TestRawString.buildRawStringVariants(constStr);
                    for (int i = 0; i < constRStrings.length; i++) {
                        assertTrue(constsParser.parseConstant(value, constRStrings[i]));
                        assertSame(value.getType(), Type.FLOAT);
                        checkParsedValue(expectedValue, value.getFloat());
                    }
                }
            }
        }
    }

    /**
     * Tests {@link FloatConstantsParser#parseConstant(Value, ReadableRawString)} method with
     * float constants containing only integer and exponent parts:
     * [<sign>]<digits>e[<sign>]<digits>.
     */
    public void testParseConstantExponentForm() {
        Value value = new Value(0);
        for (int si = 0; si < signStrings.length; si++) {
            for (int ii = 0; ii < digitsStrings.length; ii++) {
                for (int ei = 0; ei < digitsStrings.length; ei++) {
                    for (int esi = 0; esi < signStrings.length; esi++) {
                        String constStr = signStrings[si] + digitsStrings[ii]
                                          + "e" + signStrings[esi] + digitsStrings[ei];
                        double expectedValue = Double.parseDouble(constStr);
                        RawString[] constRStrings =
                            TestRawString.buildRawStringVariants(constStr);
                        for (int i = 0; i < constRStrings.length; i++) {
                            assertTrue(constsParser.parseConstant(value, constRStrings[i]));
                            assertSame(value.getType(), Type.FLOAT);
                            checkParsedValue(expectedValue, value.getFloat());
                        }
                    }
                }
            }
        }
    }

    /**
     * Tests {@link FloatConstantsParser#parseConstant(Value, ReadableRawString)} method with
     * full form of float constants: [<sign>]<digits>.<digits>e[<sign>]<digits>.
     */
    public void testParseConstantFullForm() {
        Value value = new Value(0);
        for (int si = 0; si < signStrings.length; si++) {
            for (int ii = 0; ii < digitsStrings.length; ii++) {
                for (int fi = 0; fi < digitsStrings.length; fi++) {
                    for (int ei = 0; ei < digitsStrings.length; ei++) {
                        for (int esi = 0; esi < signStrings.length; esi++) {
                            String constStr = signStrings[si] + digitsStrings[ii]
                                              + "." + digitsStrings[ii]
                                              + "e" + signStrings[esi] + digitsStrings[ei];
                            double expectedValue = Double.parseDouble(constStr);
                            RawString[] constRStrings =
                                TestRawString.buildRawStringVariants(constStr);
                            for (int i = 0; i < constRStrings.length; i++) {
                                assertTrue(constsParser.parseConstant(value, constRStrings[i]));
                                assertSame(value.getType(), Type.FLOAT);
                                checkParsedValue(expectedValue, value.getFloat());
                            }
                        }
                    }
                }
            }
        }
    }

}
