/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.Arrays;
import junit.framework.TestCase;


/**
 * Tests for {@link IntFormatter} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestIntFormatter extends TestCase {

    /**
     * Empty default constructor
     */
    public TestIntFormatter() { }

    /**
     * Tests {@link IntFormatter#formatToArray(int, char[], int)} and
     * {@link IntFormatter#getCharactersCount(int)} methods.
     * Calls the methods with different buffer sizes, start positions and number values
     * distributed over range of <code>int</code> values. Results are compared with
     * strings returned by <code>Integer.toString</code> method.
     */
    public void testFormatToArrayAndGetCharactersCount() {
        int[] specialNumbers = {
            Integer.MIN_VALUE, -1234567890, -987654321, -23456789,
            -8765432, -345678, -76543, -4567, -654, -56, -5, 0,
            1, 98, 234, 8765, 34567, 765432, 4567890, 65432109, 567890123, Integer.MAX_VALUE};
        for (int bufSize = 0; bufSize <= IntFormatter.MAX_CHARACTERS_COUNT * 2; bufSize++) {
            char[] buf = new char[bufSize];
            for (int startPos = 0; startPos <= bufSize; startPos++) {
                for (int i = 0; i < specialNumbers.length; i++) {
                    for (int delta = -10; delta <= 10; delta++) {
                        checkFormatter(specialNumbers[i] + delta, buf, startPos, '\0');
                        checkFormatter(specialNumbers[i] + delta, buf, startPos, ' ');
                    }
                }
            }
        }
    }

    /**
     * Tests the formatter against given value and output buffer, filling the whole buffer with
     * specified filler before the operation and checking whether characters before and after
     * formatters output hasn't been changed.
     *
     * @param value integer value to format
     * @param buf output buffer for formatted value
     * @param startPos index of first buffer element to use for formatted value
     * @param filler character used to fill the whole buffer before launching formatter
     */
    private void checkFormatter(int value, char[] buf, int startPos, char filler) {
        Arrays.fill(buf, filler);
        String strValue = Integer.toString(value);
        assertEquals(IntFormatter.getCharactersCount(value), strValue.length());
        int charsCount = IntFormatter.formatToArray(value, buf, startPos);
        if (startPos + strValue.length() <= buf.length) {
            assertEquals(charsCount, strValue.length());
            assertEquals(String.copyValueOf(buf, startPos, charsCount), strValue);
            checkFiller(buf, 0, startPos, filler);
            checkFiller(buf, startPos + charsCount, buf.length, filler);
        } else {
            assertEquals(charsCount, 0);
            checkFiller(buf, 0, buf.length, filler);
        }
    }

    /**
     * Checks whether specified range of components of given character array contains the same
     * given character.
     *
     * @param buf character buffer to check
     * @param startPos index of first buffer element to check
     * @param endPos index after the last buffer element to check
     * @param filler expected value of buffer elements
     */
    private void checkFiller(char[] buf, int startPos, int endPos, char filler) {
        for (int i = startPos; i < endPos; i++) {
            assertEquals(buf[i], filler);
        }
    }

}
