/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import junit.framework.TestCase;


/**
 * Tests for {@link ObjectsPool} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestObjectsPool extends TestCase {

    /**
     * A factory used to create objects for the pool and to count them.
     */
    private CountingObjectsFactory factory;

    /**
     * {@inheritDoc}
     */
    protected void setUp() throws Exception {
        super.setUp();
        factory = new CountingObjectsFactory();
    }

    /**
     * Empty default constructor
     */
    public TestObjectsPool() { }

    /**
     * Tests whether a number of objects created during pool initialization is valid.
     */
    public void testConstructor() {
        int expectedInstancesCount = 0;
        for (int size = 0; size < 1000; size += 10) {
            new ObjectsPool(factory, size);
            expectedInstancesCount += size;
            assertEquals(factory.getInstancesCount(), expectedInstancesCount);
        }
    }

    /**
     * Tests {@link ObjectsPool#ObjectsPool(IObjectsFactory, int)} constructor with invalid
     * <code>size</code> argument.
     */
    public void testConstructorInvalidArgs() {
        try {
            new ObjectsPool(factory, -1);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
        try {
            new ObjectsPool(factory, Integer.MIN_VALUE);
            fail();
        } catch (IllegalArgumentException e) {
            // OK
        }
    }

    /**
     * Gets specified number of objects from given pool.
     *
     * @param pool objects pool to use
     * @param count number of objects to get
     * @return a list of object acquired from the pool
     */
    private List getObjectsFromPool(ObjectsPool pool, int count) {
        List result = new LinkedList();
        for (int i = 0; i < count; i++) {
            Object o = pool.getObject();
            assertNotNull(o);
            result.add(o);
        }
        return result;
    }

    /**
     * Releases given objects back into given pool.
     *
     * @param pool pool to release given objects to
     * @param objects a list of objects to release to the pool
     */
    private void releaseObjectsToPool(ObjectsPool pool, List objects) {
        for (Iterator iter = objects.iterator(); iter.hasNext();) {
            Object o = iter.next();
            pool.releaseObject(o);
        }
    }

    /**
     * Tests whether a pool doesn't allocate any new objects during getting and releasing
     * objects if total number of acquired objects doesn't exceed the pool's size.
     */
    public void testGetReleaseInLimit() {
        final int size = 10;
        ObjectsPool pool = new ObjectsPool(factory, size);
        assertEquals(factory.getInstancesCount(), size);
        int aGroupSize = size / 2;
        int bGroupSize = size - aGroupSize;
        List aGroup = getObjectsFromPool(pool, aGroupSize);
        List bGroup = getObjectsFromPool(pool, bGroupSize);
        releaseObjectsToPool(pool, aGroup);
        Object aObject = pool.getObject();
        assertNotNull(aObject);
        releaseObjectsToPool(pool, bGroup);
        Object bObject = pool.getObject();
        assertNotNull(bObject);
        releaseObjectsToPool(pool, aGroup);
        pool.releaseObject(aObject);
        pool.releaseObject(bObject);
        assertEquals(factory.getInstancesCount(), size);
    }

    /**
     * Tests whether a pool allocates extra objects after exceeding its size and doesn't grow
     * further after releasing some objects and getting new ones without going beyond the size
     * of previous allocation.
     */
    public void testGetReleaseLimitExceeded() {
        final int size = 10;
        ObjectsPool pool = new ObjectsPool(factory, size);
        List group = getObjectsFromPool(pool, size);
        Object o = pool.getObject();
        assertEquals(factory.getInstancesCount(), size + 1);
        releaseObjectsToPool(pool, group);
        pool.releaseObject(o);
        getObjectsFromPool(pool, size);
        pool.getObject();
        assertEquals(factory.getInstancesCount(), size + 1);
    }

    /**
     * Tests whether after calling {@link ObjectsPool#reset()} a pool returns to its initial
     * state, i.e. offers some number of objects without creating new instances. Only
     * {@link ObjectsPool#getObject()} operations will be performed before resetting the pool.
     */
    public void testReset() {
        final int size = 10;
        ObjectsPool pool = new ObjectsPool(factory, size);
        getObjectsFromPool(pool, size + 1);
        assertEquals(factory.getInstancesCount(), size + 1);
        pool.reset();
        getObjectsFromPool(pool, size);
        assertEquals(factory.getInstancesCount(), size + 1);
        pool.getObject();
        assertEquals(factory.getInstancesCount(), size + 2);
    }

    /**
     * Tests whether after calling {@link ObjectsPool#reset()} a pool returns to its initial
     * state, i.e. offers some number of objects without creating new instances. Both
     * {@link ObjectsPool#getObject()} and {@link ObjectsPool#releaseObject(Object)} operations
     * will be performed before resetting the pool.
     */
    public void testResetAfterGetRelease() {
        final int size = 10;
        ObjectsPool pool = new ObjectsPool(factory, size);
        List aGroup = getObjectsFromPool(pool, size / 3);
        List bGroup = getObjectsFromPool(pool, size / 3);
        getObjectsFromPool(pool, size / 3);
        releaseObjectsToPool(pool, bGroup);
        releaseObjectsToPool(pool, aGroup);
        pool.reset();
        getObjectsFromPool(pool, size);
        assertEquals(factory.getInstancesCount(), size);
    }

}
