/*
 * Tests for Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import junit.framework.TestCase;


/**
 * Tests for {@link RawString} class.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class TestRawString extends TestCase {

    /**
     * Empty default constructor
     */
    public TestRawString() { }

    /**
     * Builds new {@link RawString} instance with value equal to given string, allocating
     * storage with size optionally bigger that required, filling unused elements with specified
     * padding character.
     *
     * @param str string value to build equivalent {@link RawString} instance for
     * @param leftPad
     *     number of unused characters in allocated storage before characters representing the
     *     string's value
     * @param rightPad
     *     number of unused characters in allocated storage after characters representing the
     *     string's value
     * @param
     *     padChar character to use for filling all unused storage elements (both left and right
     *     padding)
     * @return newly created {@link RawString} instance with value equal to <code>str</code>
     */
    public static RawString buildRawStringVariant(
        String str, int leftPad, int rightPad, char padChar) {
        char[] chars = new char[leftPad + str.length() + rightPad];
        Arrays.fill(chars, padChar);
        str.getChars(0, str.length(), chars, leftPad);
        return new RawString(chars, leftPad, str.length());
    }

    /**
     * Builds an array of various {@link RawString} instances with values equal to given string.
     * Returned instances will differ with size of storage and its layout, but will have the
     * same string value.
     * <p>
     * This method should be used for testing methods that accepts parameters of type
     * {@link pl.chyla.wyjec.utils.ReadableRawString} to make sure, that they don't make wrong
     * assumptions about internal layout of characters inside string's storage. When creating
     * {@link RawString} instances using {@link RawString#RawString(String)} constructor,
     * characters representing string's value always start at position 0 and the size of storage
     * is equal to string's length. Correct code should not depend on such (typical) layout.
     *
     * @param str string value to build equivalent instances of {@link RawString} for
     * @return
     *     array of equal {@link RawString} instances with different storage sizes and
     *     characters layouts, all equal to <code>str</code>
     */
    public static RawString[] buildRawStringVariants(String str) {
        List rstrings = new LinkedList();
        rstrings.add(new RawString(str));
        char[] padChars;
        if (str.length() > 0) {
            padChars = new char[] {'\0', ' ', str.charAt(0), str.charAt(str.length() - 1)};
        } else {
            padChars = new char[] {'\0', ' '};
        }
        int[] pads = {0, 1, str.length()};
        for (int li = 0; li < pads.length; li++) {
            for (int ri = 0; ri < pads.length; ri++) {
                for (int pi = 0; pi < padChars.length; pi++) {
                    rstrings.add(buildRawStringVariant(str, pads[li], pads[ri], padChars[pi]));
                }
            }
        }
        return (RawString[]) rstrings.toArray(new RawString[0]);
    }

    /**
     * Builds various {@link RawString} variants for all given strings
     * (using {@link #buildRawStringVariants(String))} and returns all generated instances in
     * single array.
     *
     * @param strings array of strings to generate {@link RawString} variants for
     * @return array of generated {@link RawString} instances
     */
    public static RawString[] buildRawStringsVariants(String[] strings) {
        List allRawStrings = new LinkedList();
        for (int i = 0; i < strings.length; i++) {
            RawString[] rstrings = buildRawStringVariants(strings[i]);
            for (int j = 0; j < rstrings.length; j++) {
                allRawStrings.add(rstrings[j]);
            }
        }
        return (RawString[]) allRawStrings.toArray(new RawString[0]);
    }

    /**
     * Generates array of <code>String</code> instances for use in tests.
     *
     * @return array of <code>String</code> instances
     */
    static String[] generateSampleStrings() {
        return new String[] {"", "x", "ab", "lizak"};
    }

    /**
     * Generates <code>String</code> instance for use in tests.
     *
     * @return sample <code>String</code> instance
     */
    static String generateSampleString() {
        return "napis";
    }

    /**
     * Generates array of <code>RawStrings</code> instances. Some of them are equal but have
     * different representation, some are identical, others are completely different.
     *
     * @return sequence of <code>RawStrings</code> instances
     */
    static RawString[] generateSampleRawStrings() {
        String[] strings = generateSampleStrings();
        List rawStrings = new LinkedList();
        for (int si = 0; si < strings.length; si++) {
            String str = strings[si];
            for (int startPos = 0; startPos <= str.length(); startPos++) {
                for (int length = 0; length <= str.length() - startPos; length++) {
                    rawStrings.add(new RawString(str.toCharArray(), startPos, length));
                }
            }
            RawString[] extraRawStrings = buildRawStringVariants(str);
            for (int i = 0; i < extraRawStrings.length; i++) {
                rawStrings.add(extraRawStrings[i]);
            }
        }
        return (RawString[]) rawStrings.toArray(new RawString[0]);
    }

    /**
     * Performs simple consistency checks for <code>RawString</code> instance.
     *
     * @param rstr <code>RawString</code> instance to check
     */
    private void verifyRawStringConsistency(RawString rstr) {
        assertEquals(rstr.getLength(), rstr.getEndPos() - rstr.getStartPos());
        assertEquals(rstr.getFreeSpace(), rstr.getStorage().length - rstr.getEndPos());
        assertEquals(rstr.toString().length(), rstr.getLength());
    }

    /**
     * Tests {@link RawString#RawString(String)} constructor.
     */
    public void testCreateFromString() {
        String[] strings = generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            String str = strings[si];
            RawString rstr = new RawString(str);
            assertEquals(rstr.toString(), str);
            assertEquals(rstr.getStorage().length, str.length());
            verifyRawStringConsistency(rstr);
        }
    }

    /**
     * Tests {@link RawString#RawString()} constructor.
     */
    public void testCreateDefault() {
        RawString rstr = new RawString();
        assertEquals(rstr.toString(), "");
        assertEquals(rstr.getStorage().length, 0);
        verifyRawStringConsistency(rstr);
    }

    /**
     * Tests {@link RawString#RawString(int)} constructor.
     */
    public void testCreateWithSize() {
        for (int i = 0; i < 10; i++) {
            RawString rstr = new RawString(i);
            assertEquals(rstr.toString(), "");
            assertEquals(rstr.getStorage().length, i);
            assertEquals(rstr.getStartPos(), 0);
            verifyRawStringConsistency(rstr);
        }
    }

    /**
     * Tests {@link RawString#RawString(int)} constructor with negative <code>size</code>
     * parameter.
     */
    public void testCreateWithSizeNegative() {
        for (int i = 1; i < 10; i++) {
            try {
                new RawString(-i);
                fail();
            } catch (IllegalArgumentException e) {
                // OK
            }
        }
    }

    /**
     * Tests {@link RawString#RawString(char[])} constructor.
     */
    public void testCreateFromArray() {
        String[] strings = generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            char[] buf = strings[si].toCharArray();
            RawString rstr = new RawString(buf);
            assertEquals(rstr.toString(), new String(buf));
            assertSame(rstr.getStorage(), buf);
            verifyRawStringConsistency(rstr);
        }
    }

    /**
     * Tests {@link RawString#RawString(char[], int, int)} constructor.
     */
    public void testCreateFromArrayPart() {
        String str = generateSampleString();
        char[] buf = str.toCharArray();
        for (int startPos = 0; startPos <= str.length(); startPos++) {
            for (int length = 0; length <= str.length() - startPos; length++) {
                RawString rstr = new RawString(buf, startPos, length);
                assertEquals(rstr.toString(), new String(buf, startPos, length));
                assertSame(rstr.getStorage(), buf);
                assertEquals(rstr.getStartPos(), startPos);
                verifyRawStringConsistency(rstr);
            }
        }
    }

    /**
     * Tests {@link RawString#RawString(char[], int, int)} constructor with invalid
     * <code>startPos</code> and <code>length</code> arguments.
     */
    public void testCreateFromArrayPartInvalidArgs() {
        String str = generateSampleString();
        char[] buf = str.toCharArray();
        for (int pos = 0; pos <= str.length(); pos++) {
            for (int i = 1; i < 10; i++) {
                try {
                    new RawString(buf, -i, pos);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    new RawString(buf, pos, -i);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    new RawString(buf, pos, str.length() - pos + i);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
            }
        }
    }

    /**
     * Tests {@link RawString#RawString(ReadableRawString)} constructor.
     */
    public void testCreateFromRawString() {
        String[] strings = generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            RawString other = new RawString(strings[si]);
            RawString rstr = new RawString(other);
            assertEquals(rstr, other);
            assertSame(rstr.getStorage(), other.getStorage());
            verifyRawStringConsistency(rstr);
        }
    }

    /**
     * Tests {@link RawString#set(char[])} method.
     */
    public void testSetArray() {
        String[] strings = generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            RawString[] rawStrings = generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                char[] buf = strings[si].toCharArray();
                RawString rstr = rawStrings[rsi];
                rstr.set(buf);
                assertEquals(rstr.toString(), new String(buf));
                assertSame(rstr.getStorage(), buf);
                verifyRawStringConsistency(rstr);
            }
        }
    }

    /**
     * Tests {@link RawString#set(char[], int, int)} method.
     */
    public void testSetArrayPart() {
        String str = generateSampleString();
        char[] buf = str.toCharArray();
        for (int startPos = 0; startPos <= str.length(); startPos++) {
            for (int endPos = startPos; endPos <= str.length(); endPos++) {
                RawString[] rawStrings = generateSampleRawStrings();
                for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                    RawString rstr = rawStrings[rsi];
                    int length = endPos - startPos;
                    rstr.set(buf, startPos, length);
                    assertEquals(rstr.toString(), new String(buf, startPos, length));
                    assertSame(rstr.getStorage(), buf);
                    assertEquals(rstr.getStartPos(), startPos);
                    verifyRawStringConsistency(rstr);
                }
            }
        }
    }

    /**
     * Tests {@link RawString#set(char[], int, int)} method with invalid
     * <code>newStartPos</code> and <code>newLength</code> arguments.
     */
    public void testSetArrayPartInvalidArgs() {
        String str = generateSampleString();
        char[] buf = str.toCharArray();
        for (int pos = 0; pos <= str.length(); pos++) {
            for (int i = 1; i < 10; i++) {
                RawString[] rawStrings = generateSampleRawStrings();
                for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                    RawString rstr = rawStrings[rsi];
                    try {
                        rstr.set(buf, -i, pos);
                        fail();
                    } catch (IllegalArgumentException e) {
                        // OK
                    }
                    try {
                        rstr.set(buf, pos, -i);
                        fail();
                    } catch (IllegalArgumentException e) {
                        // OK
                    }
                    try {
                        rstr.set(buf, pos, str.length() - pos + i);
                        fail();
                    } catch (IllegalArgumentException e) {
                          // OK
                    }
                    verifyRawStringConsistency(rstr);
                }
            }
        }
    }

    /**
     * Tests {@link RawString#set(ReadableRawString)} method,
     */
    public void testSetRawString() {
        RawString[] otherRawStrings = generateSampleRawStrings();
        for (int rsi = 0; rsi < otherRawStrings.length; rsi++) {
            RawString other = otherRawStrings[rsi];
            RawString[] rawStrings = generateSampleRawStrings();
            for (int rsj = 0; rsj < rawStrings.length; rsj++) {
                RawString rstr = rawStrings[rsj];
                rstr.set(other);
                assertEquals(rstr, other);
                assertSame(rstr.getStorage(), other.getStorage());
                verifyRawStringConsistency(rstr);
            }
        }
    }

    /**
     * Tests {@link RawString#setRange(int, int)} method.
     */
    public void testSetRange() {
        char[] buf = generateSampleString().toCharArray();
        for (int startPos = 0; startPos <= buf.length; startPos++) {
            for (int endPos = startPos; endPos <= buf.length; endPos++) {
                RawString rstr = new RawString(buf);
                int length = endPos - startPos;
                rstr.setRange(startPos, endPos - startPos);
                rstr.set(buf, startPos, length);
                assertEquals(rstr.toString(), new String(buf, startPos, length));
                assertSame(rstr.getStorage(), buf);
                assertEquals(rstr.getStartPos(), startPos);
                verifyRawStringConsistency(rstr);
            }
        }
    }

    /**
     * Tests {@link RawString#setRange(int, int)} method with invalid <code>newStartPos</code>
     * and <code>newLength</code> arguments.
     */
    public void testSetRangeInvalidArgs() {
        char[] buf = generateSampleString().toCharArray();
        for (int pos = 0; pos <= buf.length; pos++) {
            for (int i = 1; i <= 10; i++) {
                RawString rstr = new RawString(buf);
                try {
                    rstr.setRange(-i, pos);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
            }
            for (int i = 1; i <= 10; i++) {
                RawString rstr = new RawString(buf);
                try {
                    rstr.setRange(pos, buf.length - pos + i);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
            }
            for (int startPos = pos + 1; startPos <= buf.length; startPos++) {
                RawString rstr = new RawString(buf);
                try {
                    rstr.setRange(startPos, pos - startPos);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(char)} method.
     */
    public void testAppendChar() {
        char[] chars = generateSampleString().toCharArray();
        for (int ci = 0; ci < chars.length; ci++) {
            char extraChar = chars[ci];
            RawString[] rawStrings = generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                RawString rstr = rawStrings[rsi];
                String oldStr = rstr.toString();
                if (rstr.getFreeSpace() >= 1) {
                    rstr.append(extraChar);
                    String newStr = oldStr + extraChar;
                    assertEquals(rstr.toString(), newStr);
                    verifyRawStringConsistency(rstr);
                } else {
                    try {
                        rstr.append(extraChar);
                        fail();
                    } catch (RawString.RawStringOverflowException e) {
                        assertSame(e.getOverflowedString(), rstr);
                        assertEquals(
                            new String(e.getExtraCharacters()), Character.toString(extraChar));
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(char[])} method.
     */
    public void testAppendArray() {
        String[] strings = generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            char[] extraChars = strings[si].toCharArray();
            RawString[] rawStrings = generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                RawString rstr = rawStrings[rsi];
                String oldStr = rstr.toString();
                if (rstr.getFreeSpace() >= extraChars.length) {
                    rstr.append(extraChars);
                    String newStr = oldStr + new String(extraChars);
                    assertEquals(rstr.toString(), newStr);
                    verifyRawStringConsistency(rstr);
                } else {
                    try {
                        rstr.append(extraChars);
                        fail();
                    } catch (RawString.RawStringOverflowException e) {
                        assertSame(e.getOverflowedString(), rstr);
                        assertEquals(
                            new String(e.getExtraCharacters()), new String(extraChars));
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(char[], int, int)} method.
     */
    public void testAppendArrayRange() {
        String[] strings = generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            char[] extraChars = strings[si].toCharArray();
            for (int startPos = 0; startPos <= extraChars.length; startPos++) {
                for (int len = 0; len <= extraChars.length - startPos; len++) {
                    RawString[] rawStrings = generateSampleRawStrings();
                    for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                        RawString rstr = rawStrings[rsi];
                        String oldStr = rstr.toString();
                        if (rstr.getFreeSpace() >= len) {
                            rstr.append(extraChars, startPos, len);
                            String newStr = oldStr + new String(extraChars, startPos, len);
                            assertEquals(rstr.toString(), newStr);
                            verifyRawStringConsistency(rstr);
                        } else {
                            try {
                                rstr.append(extraChars, startPos, len);
                                fail();
                            } catch (RawString.RawStringOverflowException e) {
                                assertSame(e.getOverflowedString(), rstr);
                                assertEquals(
                                    new String(e.getExtraCharacters()),
                                    new String(extraChars, startPos, len));
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(char[], int, int)} method with invalid <code>start</code>
     * or <code>length</code> arguments.
     */
    public void testAppendArrayRangeInvalidArgs() {
        RawString[] rawStrings = generateSampleRawStrings();
        char[] extraChars = "abc".toCharArray();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            RawString rstr = rawStrings[rsi];
            try {
                rstr.append(extraChars, -1, 0);
                fail();
            } catch (IllegalArgumentException e) {
                // OK
            }
            try {
                rstr.append(extraChars, 0, -1);
                fail();
            } catch (IllegalArgumentException e) {
                // OK
            }
            try {
                rstr.append(extraChars, 1, extraChars.length);
                fail();
            } catch (IllegalArgumentException e) {
                // OK
            }
            try {
                rstr.append(extraChars, extraChars.length, 1);
                fail();
            } catch (IllegalArgumentException e) {
                // OK
            }
        }
    }

    /**
     * Tests {@link RawString#append(ReadableRawString)} method.
     */
    public void testAppendRawString() {
        RawString[] extraRawStrings = generateSampleRawStrings();
        for (int ersi = 0; ersi < extraRawStrings.length; ersi++) {
            RawString extraRawString = extraRawStrings[ersi];
            RawString[] rawStrings = generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                RawString rstr = rawStrings[rsi];
                String oldStr = rstr.toString();
                if (rstr.getFreeSpace() >= extraRawString.getLength()) {
                    rstr.append(extraRawString);
                    String newStr = oldStr + extraRawString.toString();
                    assertEquals(rstr.toString(), newStr);
                    verifyRawStringConsistency(rstr);
                } else {
                    try {
                        rstr.append(extraRawString);
                        fail();
                    } catch (RawString.RawStringOverflowException e) {
                        assertSame(e.getOverflowedString(), rstr);
                        assertEquals(
                            new String(e.getExtraCharacters()), extraRawString.toString());
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(ReadableRawString, int, int)} method.
     */
    public void testAppendRawStringRange() {
        RawString[] extraRawStrings = generateSampleRawStrings();
        for (int ersi = 0; ersi < extraRawStrings.length; ersi++) {
            RawString extraRawString = extraRawStrings[ersi];
            for (int startPos = 0; startPos <= extraRawString.getLength(); startPos++) {
                for (int len = 0; len <= extraRawString.getLength() - startPos; len++) {
                    RawString[] rawStrings = generateSampleRawStrings();
                    for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                        RawString rstr = rawStrings[rsi];
                        String oldStr = rstr.toString();
                        if (rstr.getFreeSpace() >= len) {
                            rstr.append(extraRawString, startPos, len);
                            String extraSubStr =
                                extraRawString.toString().substring(startPos, startPos + len);
                            assertEquals(rstr.toString(), oldStr + extraSubStr);
                            verifyRawStringConsistency(rstr);
                        } else {
                            try {
                                rstr.append(extraRawString, startPos, len);
                                fail();
                            } catch (RawString.RawStringOverflowException e) {
                                assertSame(e.getOverflowedString(), rstr);
                                assertEquals(
                                    new String(e.getExtraCharacters()),
                                    extraRawString.toString().substring(
                                        startPos, startPos + len));
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(ReadableRawString, int, int)} method with invalid
     * <code>start</code> or <code>length</code> arguments.
     */
    public void testAppendRawStringRangeInvalidArgs() {
        RawString[] rawStrings = generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            RawString rstr = rawStrings[rsi];
            RawString[] extraRawStrings = generateSampleRawStrings();
            for (int ersi = 0; ersi < extraRawStrings.length; ersi++) {
                RawString extraRawString = extraRawStrings[ersi];
                try {
                    rstr.append(extraRawString, -1, 0);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    rstr.append(extraRawString, 0, -1);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    rstr.append(extraRawString, 1, extraRawString.getLength());
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    rstr.append(extraRawString, extraRawString.getLength(), 1);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(String)} method.
     */
    public void testAppendString() {
        String[] strings = generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            String extraString = strings[si];
            RawString[] rawStrings = generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                RawString rstr = rawStrings[rsi];
                String oldStr = rstr.toString();
                if (rstr.getFreeSpace() >= extraString.length()) {
                    rstr.append(extraString);
                    String newStr = oldStr + extraString;
                    assertEquals(rstr.toString(), newStr);
                    verifyRawStringConsistency(rstr);
                } else {
                    try {
                        rstr.append(extraString);
                        fail();
                    } catch (RawString.RawStringOverflowException e) {
                        assertSame(e.getOverflowedString(), rstr);
                        assertEquals(new String(e.getExtraCharacters()), extraString);
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(String, int, int)} method.
     */
    public void testAppendStringRange() {
        String[] extraStrings = generateSampleStrings();
        for (int esi = 0; esi < extraStrings.length; esi++) {
            String extraString = extraStrings[esi];
            for (int startPos = 0; startPos <= extraString.length(); startPos++) {
                for (int len = 0; len <= extraString.length() - startPos; len++) {
                    RawString[] rawStrings = generateSampleRawStrings();
                    for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                        RawString rstr = rawStrings[rsi];
                        String oldStr = rstr.toString();
                        if (rstr.getFreeSpace() >= len) {
                            rstr.append(extraString, startPos, len);
                            String extraSubStr =
                                extraString.substring(startPos, startPos + len);
                            assertEquals(rstr.toString(), oldStr + extraSubStr);
                            verifyRawStringConsistency(rstr);
                        } else {
                            try {
                                rstr.append(extraString, startPos, len);
                                fail();
                            } catch (RawString.RawStringOverflowException e) {
                                assertSame(e.getOverflowedString(), rstr);
                                assertEquals(
                                    new String(e.getExtraCharacters()),
                                    extraString.substring(startPos, startPos + len));
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(String, int, int)} method with invalid <code>start</code>
     * or <code>length</code> arguments.
     */
    public void testAppendStringRangeInvalidArgs() {
        RawString[] rawStrings = generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            RawString rstr = rawStrings[rsi];
            String[] extraStrings = generateSampleStrings();
            for (int esi = 0; esi < extraStrings.length; esi++) {
                String extraString = extraStrings[esi];
                try {
                    rstr.append(extraString, -1, 0);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    rstr.append(extraString, 0, -1);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    rstr.append(extraString, 1, extraString.length());
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
                try {
                    rstr.append(extraString, extraString.length(), 1);
                    fail();
                } catch (IllegalArgumentException e) {
                    // OK
                }
            }
        }
    }

    /**
     * Tests {@link RawString#append(StringBuffer)} method.
     */
    public void testAppendStringBuffer() {
        StringBuffer sbuf = new StringBuffer();
        String[] strings = generateSampleStrings();
        for (int si = 0; si < strings.length; si++) {
            sbuf.setLength(0);
            sbuf.append(strings[si]);
            RawString[] rawStrings = generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                RawString rstr = rawStrings[rsi];
                String oldStr = rstr.toString();
                if (rstr.getFreeSpace() >= sbuf.length()) {
                    rstr.append(sbuf);
                    String newStr = oldStr + sbuf.toString();
                    assertEquals(rstr.toString(), newStr);
                    verifyRawStringConsistency(rstr);
                } else {
                    try {
                        rstr.append(sbuf);
                        fail();
                    } catch (RawString.RawStringOverflowException e) {
                        assertSame(e.getOverflowedString(), rstr);
                        assertEquals(new String(e.getExtraCharacters()), sbuf.toString());
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#extend(int)} method.
     */
    public void testExtend() {
        for (int extraCharsCount = 0; extraCharsCount < 20; extraCharsCount++) {
            RawString[] rawStrings = generateSampleRawStrings();
            for (int rsi = 0; rsi < rawStrings.length; rsi++) {
                RawString rstr = rawStrings[rsi];
                String oldStr = rstr.toString();
                if (rstr.getFreeSpace() >= extraCharsCount) {
                    rstr.extend(extraCharsCount);
                    assertEquals(rstr.getLength(), oldStr.length() + extraCharsCount);
                    assertEquals(rstr.toString().substring(0, oldStr.length()), oldStr);
                    verifyRawStringConsistency(rstr);
                } else {
                    try {
                        rstr.extend(extraCharsCount);
                        fail();
                    } catch (RawString.RawStringOverflowException e) {
                        assertSame(e.getOverflowedString(), rstr);
                    }
                }
            }
        }
    }

    /**
     * Tests {@link RawString#extend(int)} method with invalid <code>extensionAmount</code>
     * argument.
     */
    public void testExtendInvalidArgs() {
        RawString[] rawStrings = generateSampleRawStrings();
        for (int rsi = 0; rsi < rawStrings.length; rsi++) {
            try {
                rawStrings[rsi].extend(-1);
                fail();
            } catch (IllegalArgumentException e) {
                // OK
            }
        }
    }

}
